#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains base classes for Qt models
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *

import tpDcc as tp
from tpDcc.libs.qt.core import qtutils


class ItemSelection(QItemSelection):
    """
    Extends QItemSelection functionality for view items
    """

    def __init__(self, *args):
        super(ItemSelection, self).__init__(*args)

    def indexes(self):
        """
        Override QItemSelection indexes() method

        :warning: Specific to Qt 4.7.0 the QModelIndexList destructor will cause a destructor crash.
            This method avoids that crash by overriding the indexes() method and manually building
            the index list with a python builtin list.
        :seealso: http://www.qtcentre.org/threads/16933
        :return: list<QModelIndex>, list of model indexes corresponding to the selected items
        """

        indexes = list()
        for i in range(self.count()):
            parent = self[i].parent()
            for c in range(self[i].left(), self[i].right() + 1):
                for r in range(self[i].top(), self[i].bottom() + 1):
                    indexes.append(self[i].model().index(r, c, parent))

        return indexes


class ListModel(QAbstractListModel, object):
    def __init__(self, data=None, parent=None):
        """
        Basic model for string lists
        :param data: list<string>, list of string items to add to the model
        :param parent: QWidget
        """

        if data is None:
            data = list()

        super(ListModel, self).__init__(parent=parent)
        self._items = data

    def headerData(self):
        return None

    def rowCount(self, parent=QModelIndex()):
        """
        Returns the length of the internal collection item list
        """

        return len(self._items)

    def columnCount(self, parent=QModelIndex()):
        """
        Returns the number of columns in this model
        """

        return 1

    def flags(self, index):
        """
        Gets the Qt.ItemFlags for the model data at a given index
        :param index: int, lookup key for the data
        :return: A valid combination of the Qt.Flags enum
        """

        # return Qt.ItemIsEditable | Qt.ItemIsEnabled | Qt.ItemIsSelectable
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def data(self, index, role):
        if role == Qt.DisplayRole:
            return self.item(index)

    def setData(self, index, value, role=Qt.EditRole):
        """
        Sets the model data at a given index, filtered by the given role to the value
        """

        return False

    def insertRows(self, position, rows, parent=QModelIndex()):
        """
        Inserts a row with empty AbstractDataItems into the model
        """

        self.beginInsertRows(parent, position, position + rows - 1)
        for i in range(rows):
            self._items.insert(position, 'test')
        self.endInsertRows()

        return True

    def removeRows(self, position, rows, parent=QModelIndex()):
        """
        Removes a row from the model
        """

        self.beginRemoveRows(parent, position, position + rows - 1)
        for i in range(rows):
            value = self._items[position]
            self._items.remove(value)
        self.endRemoveRows()
        return True

    def clear(self):
        """
        Clears all data model
        """

        try:
            self._items.clear()
        except Exception:
            del self._items[:]

    def item(self, index):
        """
        Returns the internal data item given the index
        :param index: int, QModelIndex representing the index
        :return: item if the index is valid, None otherwise
        """

        if isinstance(index, (int, long)):
            return self._items[index]

        if isinstance(index, QModelIndex) and index.isValid():
            item = index.internalPointer()
            if item:
                return item
            return self._items[index.row()]

        return None

    def set_items(self, items):
        """
        Clears current model items and adds new ones
        :param items: list<string>, items to add to the model
        """

        self.clear()
        for item in items:
            self.append_item(item)

    def append_item(self, item):
        """
        Appends an existing AbstractDataItem into the model
        :param item: AbstractDataItem
        :return: bool
        """

        next_index = self.rowCount()
        last_index = next_index
        self.beginInsertRows(QModelIndex(), next_index, last_index)
        self._item_insert(item)
        self.endInsertRows()
        return True

    def _item_insert(self, item=None):
        """
        Internal function that inserts an item into the internal collection
        :param item: object, item to append to end of the internal collection
        """

        self._items.append(item)

    def _item_insert_position(self, item, index):
        """
        Internal item insert at specific index position
        :param item: object, item to append to end of the internal collection
        :param index: int, internal index of the item in the internal collection
        """

        self._items.insert(index, item)

    def _item_remove(self, item):
        """
        Internal item removal
        :param item: the item to remove from the internal collection
        """

        try:
            self._items.remove(item)
        except ValueError:
            pass

    def _item_remove_position(self, index):
        """
        Internal remove item at specific index position in the internal collection
        :param index: int, index of the item to remove from the internal collection
        """

        try:
            item = self._items.pop(index)
        except IndexError:
            pass


class TableModel(QAbstractTableModel, object):
    def __init__(self, data=[], horizontal_headers=[], vertical_headers=[], parent=None):
        """
        Basic model for table models
        :param data: list<list>, multi dimensional array in row-column order
        :param parent: QWidget
        """

        super(TableModel, self).__init__(parent=parent)
        self._items = data
        self._headers = {
            Qt.Horizontal: horizontal_headers,
            Qt.Vertical: vertical_headers
        }

    def headerData(self, section, orientation, role):

        if role != Qt.DisplayRole:
            return None

        header_data = self._headers.get(orientation, None)
        return header_data[section] if header_data else None

    def rowCount(self, parent=QModelIndex()):
        """
        Returns the length of the internal collection item list
        """

        return len(self._items)

    def columnCount(self, parent=QModelIndex()):
        """
        Returns the length of the first element of the internal collection item list
        """

        return len(self._headers[Qt.Horizontal])

    def flags(self, index):
        """
        Gets the Qt.ItemFlags for the model data at a given index
        :param index: int, lookup key for the data
        :return: A valid combination of the Qt.Flags enum
        """

        # return Qt.ItemIsEditable | Qt.ItemIsEnabled | Qt.ItemIsSelectable
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def data(self, index, role):

        if role == Qt.DisplayRole:
            return self.item(index=index)

    def setData(self, index, value, role=Qt.EditRole):
        """
        Sets the model data at a given index, filtered by the given role to the value
        """

        return False

    def insertRows(self, position, rows, parent=QModelIndex()):
        """
        Inserts a new item into the table
        """

        self.beginInsertRows(parent, position, position + rows - 1)
        for i in range(rows):
            default_values = ['' for i in range(self.columnCount())]
            self._items.insert(position, default_values)
        self.endInsertRows()

        return True

    def insertColumns(self, position, columns, parent=QModelIndex()):
        self.beginInsertColumns(parent, position, position + columns - 1)
        row_count = len(self._items)
        for i in range(columns):
            for j in range(row_count):
                self._items[j].insert(position, '')
        self.endInsertColumns()

        return True

    def clear(self):
        """
        Clears all data model
        """

        try:
            self._items.clear()
        except Exception:
            del self._items[:]

    def item(self, index):
        """
        Returns the internal data item given the index
        :param index: int, QModelIndex representing the index
        :return: item if the index is valid, None otherwise
        """

        if isinstance(index, (int, long)):
            return self._items[index]

        if isinstance(index, QModelIndex) and index.isValid():
            item = index.internalPointer()
            if item:
                return item

            if len(self._items) > 0:
                return self._items[index.row()][index.column()]
        return None

    def set_items(self, items):
        """
        Clears current model items and adds new ones
        :param items: list<list>, items to add to teh model
        """

        self.clear()
        for item in items:
            self.append_item(item=item)

    def append_item(self, item):
        """
        Appends an existing AbstractDataItem into the model
        :param item: AbstractDataItem
        :return: bool
        """

        self.beginInsertRows(QModelIndex(), len(self._items), len(self._items))
        self._item_insert(item)
        self.endInsertRows()

        return True

    def _item_insert(self, item):
        """
        Internal function that inserts an item into the internal collection
        :param item: object, item to append to end of the internal collection
        """

        self._items.append(item)

    def _item_insert_position(self, item, index):
        """
        Internal item insert at specific index position
        :param item: object, item to append to end of the internal collection
        :param index: int, internal index of the item in the internal collection
        """

        self._items.insert(index, item)


class BaseTreeItem(QObject):

    childAdded = Signal(object)
    childRemoved = Signal(object)

    if qtutils.is_pyside2():
        dataChanging = Signal(object, object, object)
        dataChanged = Signal(object, object, object)
    else:
        dataChanging = Signal(object, object)
        dataChanged = Signal(object, object)

    def __init__(self, data, parent=None):
        self._item_data = data or list()
        self._child_items = list()
        super(BaseTreeItem, self).__init__(parent)

    def data(self, column):
        return self._item_data[column]

    def set_data(self, column, value):
        if column <= 0 or column >= len(self._item_data):
            return False

        self._item_data[column] = value

        return True

    def row(self):
        """
        Returns the row index of this item within the parent's child collection
        :return: QModelIndex, the respective index if parent is valid; O otherwise
        """

        if self.parent():
            return self.parent().child_index(self)

        return 0

    def column(self):
        """
        Returns the column index of this item within the parent's child collection
        :return: QModelIndex, the respective index if the parent is valid; 0 otherwise
        """

        return 0

    def data(self, column):
        """
        Gets the data in the given column
        :param column: int
        :return: variant
        """

        return self._item_data[column]

    def flags(self, column):
        """
        Get the Qt.ItemFlags for the model data at a given index
        :return: A valid combination of the QtCore.Qt.QFlags enum.
        """

        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def column_count(self):
        """
         Returns number of columns stored at this item
         :return: int, number of columns at this item
         """

        return len(self._item_data)

    def row_count(self):
        """
        Returns number of rows stored at this item
        :return: int, number of rows at this item
        """

        return len(self._child_items)

    def child(self, row):
        """
        Returns the child BaseTreeItem at the given row index
        :param row: int, index into the list of children
        :return: BaseTreeItem, respective item if row is valid; None otherwise
        """

        return self._child_items[row] if 0 <= row < len(self._child_items) else None

    def has_children(self):
        """
        Returns whether or not this item has children
        :return: bool
        """

        return bool(self._child_items)

    def child_count(self):
        """
        Returns the number of children this item has
        :return: int, number of children this item has
        """

        return len(self._child_items)

    def child_index(self, child):
        """
        Returns the index of the given item in the internal list of children of this item
        :param child: BaseTreeItem, item in the list of children
        :return: int, lowest index in collection that item appears
        """

        return self._child_items.index(child)

    def is_root(self):
        """
        Returns whether or not current node is a root one (has no parents)
        :return: bool
        """

        return not bool(self.parent())

    def append_child(self, item):
        """
        Appends a child item to the internal collection of children
        :return: object, item to append
        """

        self._child_items.append(item)
        item.setParent(self)
        self.childAdded.emit(item)

    def insert_child(self, position, item):
        """
        Inserts a child item into the internal collection of children
        :param position: int, position to insert item into
        :param item: object, item to insert
        """

        if position < 0 or position > len(self._child_items):
            return False

        self._child_items.insert(position, item)
        item.setParent(self)
        self.childAdded.emit(item)

    def remove_child(self, item):
        """
        Removes an item from the children list
        :param item: object, item to remove
        """

        return self.remove_index(self.child_index(item))

    def remove(self):
        """
        Removes current item from its parent
        """

        if self.parent():
            self.parent().remove_child(self)
        for child in self._child_items:
            self.remove_child(child)

    def clear(self):
        """
        Clears all data from the item (and its children)
        """

        for child in self._child_items:
            child.clear()
        self.clear()


class TreeModel(QAbstractItemModel, object):
    def __init__(self, header_data=['']):
        self._root = self._create_root(header_data)
        super(TreeModel, self).__init__()

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def columnCount(self, parent=QModelIndex()):
        """
        Overrides columnCount base function
        Returns the number of columns in the model
        :param parent: QModelIndex
        :return: int
        """

        parent_item = self.item(parent)
        return parent_item.column_count()

    def rowCount(self, parent=QModelIndex()):
        """
        Overrides rowCount base function
        Returns the number of rows in this model
        :param parent: QModelIndex
        :return: int
        """

        parent_item = self.item(parent)
        return parent_item.row_count()

    def data(self, index, role):
        """
        Overrides base data function
        Returns the model data for the given index, filtered by the given role
        :param index: QModelIndex
        :param role: QtRole
        :return: variant
        """

        if not index.isValid():
            return None

        if role != Qt.DisplayRole and role != Qt.EditRole:
            return None

        item = self.item(index)
        return item.data(index.column()) if item else None

    def setData(self, index, value, role=Qt.EditRole):
        """
        Overrides setData base function
        Sets the model data at a given index, filtered by the given role to the value
        :param index: QModelIndex
        :param value: variant
        :param role: QtRole
        """

        if role != Qt.EditRole:
            return False

        item = self.item(index)
        if not item:
            return False

        res = item.set_data(index.column(), value)
        if not res:
            return False

        self.dataChanged.emit(index, value)

        return res

    def flags(self, index):
        """
        Overrides flags base function
        Get the Qt.ItemFlags for the model data at a given index
        :param index: int, lookup key to the data
        :return: A valid combination of the Qt.QFlags enum
        """

        if not index.isValid():
            return 0

        item = self.item(index)
        if item:
            return item.flags(index.column())

        return Qt.NoItemFlags

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """
        Overrides headerData base function
        Returns the header dat for this model
        """

        if orientation == Qt.Horizontal and role == Qt.DisplayRole and self._root:
            return self._root.data(section)

        return None

    def index(self, row, column, parent=QModelIndex()):
        """
        Returns the index of the internal data item in the model specified by the given row, column and parent index
        :param row: int, row index into the model
        :param column: int column index into the model
        :param parent: QModelIndex, option parent QModelIndex
        :return: QModelIndex, QModelIndex of the lookuo operation
        """

        if not self.hasIndex(row, column, parent) or (parent.isValid() and parent.column() != 0):
            return QModelIndex()

        parent_item = self.item(parent)
        child_item = parent_item.child(row)
        if child_item:
            return self.createIndex(row, column, child_item)
        else:
            return QModelIndex()

    def insertRows(self, position, rows, parent=QModelIndex()):
        """
        Inserts a new BaseTreeItem into the model
        :param position: int, position to insert the row
        :param rows: int, number of rows to insert
        :param parent: QModelIndex, index of the parent item in the model
        :return: bool, True if the insertion was successful; False otherwise
        """

        parent_item = self.item(parent)
        end_index = position + rows
        next_index = position
        last_index = end_index - 1
        self.beginInsertRows(parent, next_index, last_index)
        for i in range(next_index, end_index):
            self._item_insert(parent_item, self.create_item(parent_item), i)
        self.endInsertRows()

        return True

    def removeRows(self, position, rows, parent=QModelIndex()):
        """
        Removes BaseTreeItem from the model
        :param position: int, position to remove the row
        :param rows: int, number of rows to remove
        :param parent: QModelIndex, index of the parent item in the model
        :return: bool, True if the removal was successful; False otherwise
        """

        parent_item = self.item(parent)
        end_index = position + rows
        next_index = position
        last_index = end_index - 1
        self.beginRemoveRows(parent, next_index, last_index)
        for i in sorted(range(next_index, end_index), reverse=True):
            self._item_remove_position(parent_item, i)
        self.endRemoveRows()

        return True

    def insertColumns(self, position, columns, parent=QModelIndex()):
        self.beginInsertColumns(parent, position, position + columns - 1)
        success = self.rootItem.insertColumns(position, columns)
        self.endInsertColumns()

        return success

    def removeColumns(self, position, columns, parent=QModelIndex()):
        self.beginRemoveColumns(parent, position, position + columns - 1)
        success = self.rootItem.removeColumns(position, columns)
        self.endRemoveColumns()

        if self.rootItem.columnCount() == 0:
            self.removeRows(0, self.rowCount())

        return success

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def root(self):
        """
        Returns root item
        :return: BaseTreeItem
        """

        return self._root

    def create_item(self, *args):
        """
        Generates a new instance of the internal data root item wit the given arguments
        :param args: list, list or arguments to pass on the root item
        :return: BaseTreeItem, root item created
        """

        return BaseTreeItem(*args)

    def item(self, index):
        """
        Returns the internal data item given the index
        :param index: QModelIndex, QModelIndex representing the item
        :return: AbstractDataTreeItem, respective item if the index is valid; None otherwise
        """

        if index.isValid():
            item = index.internalPointer()
            if item:
                return item

        return self._root

    def item_index(self, item):
        """
        Returns the index for a given item
        :param item: BaseTreeItem, item already in the model
        :return: QModelIndex, lookup operation
        """

        if not item or item == self._root:
            return QModelIndex()

        return self.createIndex(item.row(), item.column(), item)

    def parent(self, index):
        """
        Overrides parent base function
        :param index: QModelIndex, QModelIndex to be used in lookup
        :return: QModelIndex, QModelIndex of the respective parent item
        """

        if not index.isValid():
            return QModelIndex()

        child_item = index.internalPointer()
        parent_item = child_item.parent()
        if parent_item is None or parent_item == self._root:
            return QModelIndex()

        return self.createIndex(parent_item.row(), parent_item.column(), parent_item)

    def append_item(self, item, parent=QModelIndex()):
        """
        Appends an existing AbstractDataTreeItem into the model
        :param item: AbstractDataTreeItem, item to insert
        :param parent: QModelIndex, index of the parent item in the model
        :return: bool, True if the insertion was successful; False otherwise
        """

        parent_item = self.item(parent)
        next_index = parent_item.child_count()
        last_index = next_index

        self.beginInsertRows(parent, next_index, last_index)

        self._item_insert(parent_item, item, next_index)

        self.endInsertRows()

        return self.item_index(item)

    def remove_item(self, item, parent=QModelIndex()):
        """
        Removes an existing AbstractDataTreeItem from the model
        :param item: AbstractDataTreeItem, item to remove
        :param parent: QModelIndex, index of the parent item in the model
        :return: bool, True if the removal was successful; False otherwise
        """

        return self.removeRows(item.row(), 1, parent)

    def clear(self):
        """
        Clears the model data
        """

        if self.hasChildren():
            self.removeRows(0, self.rowCount())

    def delete(self):
        """
        Prepares the model for destruction
        """

        self._root.clear()
        self.clear()

    # =================================================================================================================
    # INTERNAL
    # =================================================================================================================

    def _create_root(self, *args):
        """
        Internal function that creates a new instance of the internal data root item with the given arguments
        :param args: list, list of arguments to pass on the root item
        :return: BaseTreeItem
        """

        return BaseTreeItem(*args)

    def _item_changing(self, id, role):
        """
        Internal data item changed event handler
        """

        item = self.sender()
        index = self.createIndex(item.row(), id, item)

    def _item_changed(self, id, role):
        """
        Internal item changed event handler
        """

        item = self.sender()
        index = self.createIndex(item.row(), id, item)
        if qtutils.is_pyside2():
            self.dataChanged.emit(index, index, None)
        elif qtutils.is_pyside():
            self.dataChanged.emit(index, index)
        else:
            tp.logger.error('You have neither PySide or PySide2, that functionality is not supported!')

    def _item_append(self, parent, item):
        """
        Internal item append
        :param parent: BaseTreeItem, parent that will contain the item
        :param item: BaseTreeItem, item to append to end of the parent's internal collection
        """

        parent.append_child(item)
        item.setParent(parent)
        self._item_connect(item)

    def _item_insert(self, parent, item, position):
        """
        Internal item insert function
        :param parent: BaseTreeItem, parent that will contain the item
        :param item: BaseTreeItem, item to insert into parent's internal collection
        :param position: int, position of the item in the parent's internal collection
        """

        parent.insert_child(position, item)
        item.setParent(parent)
        self._item_connect(item)

    def _item_remove(self, parent, item):
        """
        Internal item removal function
        :param parent: AbstractDataTreeItem, parent from we want to remove item
        :param item: AbstractDataTreeItem, item we want to remove
        """

        parent.remove_child(item)
        item.setParent(None)
        self._item_disconnect(item)

    def _item_remove_position(self, parent, index):
        """
        Internal item removal at position function
        :param parent: AbstractDataTreeItem, parent from we want to remove item
        :param index: int, index of the chld we want to remove from parent's internal collection
        """

        item = parent.remove_index(index)
        item.setParent(None)
        self._item_disconnect(item)

    def _item_connect(self, item):
        """
        Internal item signal connection
        :param item: AbstractDataTreeItem, item we want to connect signals to
        """

        item.dataChanging.connect(self._item_changing)
        item.dataChanged.connect(self._item_changed)

    def _item_disconnect(self, item):
        """
        Internal item signal disconnection
        :param item: AbstractDataTreeItem, item we want to disconnect signals from
        """

        item.dataChanging.disconnect()
        item.dataChanged.disconnect()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains library viewer widget implementation
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

from tpDcc.libs import qt
from tpDcc.libs.qt.core import base
from tpDcc.libs.qt.widgets import toast, action
from tpDcc.libs.qt.widgets.library import consts, treewidget, listview, items


class LibraryViewerDelegate(QStyledItemDelegate, object):
    """
    Class that defines visual style in LibraryViewer of LibraryItems
    """

    def __init__(self):
        super(LibraryViewerDelegate, self).__init__()

        self._viewer = None

    """
    ##########################################################################################
    OVERRIDES
    ##########################################################################################
    """

    def sizeHint(self, option, index):
        """
        Overrides base QStyledItemDelegate sizeHint function
        Return the size for the given id¡ndex
        :param option: QStylOptionViewItem
        :param index: QModelIndex
        :return: QSize
        """

        item = self.viewer().item_from_index(index)
        if isinstance(item, items.LibraryGroupItem):
            return item.sizeHint()

        return self.viewer().item_size_hint()

    def paint(self, painter, option, index):
        """
        Overrides base QStyledItemDelegate paint function
        Paint performs low-level painting for the given model index
        :param painter: QPainter
        :param option: QStyleOptionViewItem
        :param index: QModelIndex
        """

        item = self.viewer().item_from_index(index)
        item.paint(painter, option, index)

    """
    ##########################################################################################
    VIEWER
    ##########################################################################################
    """

    def viewer(self):
        """
        Returns LibraryViewer object associated to this delegate
        :return: LibraryViewer
        """

        return self._viewer

    def set_viewer(self, viewer):
        """
        Set LibraryViewer associated to this delegate
        :param viewer: LibraryViewer
        """

        self._viewer = viewer


class LibraryViewer(base.BaseWidget, object):
    """
    Class that implements library viewer widget
    """

    IconMode = consts.DEFAULT_ICON_MODE
    TableMode = consts.DEFAULT_TABLE_MODE

    DEFAULT_PADDING = consts.VIEWER_DEFAULT_PADDING
    DEFAULT_ZOOM_AMOUNT = consts.VIEWER_DEFAULT_ZOOM_AMOUNT
    DEFAULT_TEXT_HEIGHT = consts.VIEWER_DEFAULT_TEXT_HEIGHT
    DEFAULT_WHEEL_SCROLL_STEP = consts.VIEWER_DEFAULT_WHEEL_SCROLL_STEP
    DEFAULT_MIN_SPACING = consts.VIEWER_DEFAULT_MIN_SPACING
    DEFAULT_MAX_SPACING = consts.VIEWER_DEFAULT_MAX_SPACING
    DEFAULT_MIN_LIST_SIZE = consts.VIEWER_DEFAULT_MIN_LIST_SIZE
    DEFAULT_MIN_ICON_SIZE = consts.VIEWER_DEFAULT_MIN_ICON_SIZE

    DEFAULT_TEXT_COLOR = consts.VIEWER_DEFAULT_TEXT_COLOR
    DEFAULT_SELECTED_TEXT_COLOR = consts.VIEWER_DEFAULT_SELECTED_TEXT_COLOR
    DEFAULT_BACKGROUND_COLOR = consts.VIEWER_DEFAULT_BACKGROUND_COLOR
    DEFAULT_BACKGORUND_HOVER_COLOR = consts.VIEWER_DEFAULT_BACKGROUND_HOVER_COLOR
    DEFAULT_BACKGROUND_SELECTED_COLOR = consts.VIEWER_DEFAULT_BACKGROUND_SELECTED_COLOR

    TREE_WIDGET_CLASS = treewidget.LibraryTreeWidget
    LIST_VIEW_CLASS = listview.LibraryListView
    DELEGATE_CLASS = LibraryViewerDelegate

    itemClicked = Signal(object)
    itemDoubleClicked = Signal(object)
    zoomChanged = Signal(object)
    spacingChanged = Signal(object)
    groupClicked = Signal(object)

    def __init__(self, parent=None):

        self._dpi = 1
        self._padding = self.DEFAULT_PADDING

        self._library = None
        self._tree_widget = None
        self._list_widget = None
        self._delegate = None
        self._is_item_text_visible = True
        self._toast_enabled = True

        self._zoom_amount = self.DEFAULT_ZOOM_AMOUNT
        self._icon_size = QSize(self._zoom_amount, self._zoom_amount)
        self._item_size_hint = QSize(self._zoom_amount, self._zoom_amount)

        self._text_color = self.DEFAULT_TEXT_COLOR
        self._text_selected_color = self.DEFAULT_SELECTED_TEXT_COLOR
        self._background_color = self.DEFAULT_BACKGROUND_COLOR
        self._background_hover_color = self.DEFAULT_BACKGORUND_HOVER_COLOR
        self._background_selected_color = self.DEFAULT_BACKGROUND_SELECTED_COLOR

        super(LibraryViewer, self).__init__(parent=parent)

    def get_main_layout(self):
        main_layout = QHBoxLayout(self)
        main_layout.setContentsMargins(0, 0, 0, 0)
        return main_layout

    def ui(self):
        super(LibraryViewer, self).ui()

        self._tree_widget = self.TREE_WIDGET_CLASS(self)

        self._list_view = self.LIST_VIEW_CLASS(self)
        self._list_view.set_tree_widget(self._tree_widget)

        self._delegate = self.DELEGATE_CLASS()
        self._delegate.set_viewer(self)
        self._list_view.setItemDelegate(self._delegate)
        self._tree_widget.setItemDelegate(self._delegate)

        self._toast_widget = toast.ToastWidget(self)
        self._toast_widget.hide()

        self.main_layout.addWidget(self._tree_widget)
        self.main_layout.addWidget(self._list_view)

        self.itemMoved = self._list_view.itemMoved
        self.itemDropped = self._list_view.itemDropped
        self.itemSelectionChanged = self._tree_widget.itemSelectionChanged

    def setup_signals(self):
        header = self.tree_widget().header()
        header.sortIndicatorChanged.connect(self._on_sort_indicator_changed)
        self._list_view.itemClicked.connect(self._on_item_clicked)
        self._list_view.itemDoubleClicked.connect(self._on_item_double_clicked)
        self._tree_widget.itemClicked.connect(self._on_item_clicked)
        self._tree_widget.itemDoubleClicked.connect(self._on_item_double_clicked)

    def wheelEvent(self, event):
        """
        Overrides base wheelEvent function
        Triggered on any wheel events for the current widget
        :param event: QWheelEvent
        """

        modifier = QApplication.keyboardModifiers()
        valid_modifiers = (Qt.AltModifier, Qt.ControlModifier)
        if modifier in valid_modifiers:
            num_degrees = event.delta() / 8
            num_steps = num_degrees / 15
            delta = (num_steps * self.wheel_scroll_step())
            value = self.zoom_amount() + delta
            self.set_zoom_amount(value)

    def contextMenuEvent(self, event):
        """
        Overrides base contextMenuEvent function
        Shows context menu for this widget
        :param event: QEvent
        """

        menu = self.create_context_menu()
        point = QCursor.pos()
        return menu.exec_(point)

    """
    ##########################################################################################
    BASE
    ##########################################################################################
    """

    def library(self):
        """
        Returns the library attached to the viewer
        :return: variant
        """

        return self._library

    def set_library(self, library):
        """
        Sets the data that will be showed in viewer
        :param library:
        """

        self._library = library
        self.set_column_labels(library.Fields)
        library.searchFinished.connect(self._on_update_items)

    def is_icon_view(self):
        """
        Returns whether widget is in icon mode or not
        :return: bool
        """

        return not self._list_view.isHidden()

    def is_table_view(self):
        """
        Returns whether widget is in list mode or not
        :return: bool
        """

        return not self._tree_widget.isHidden()

    def set_view_mode(self, mode):
        """
        Sets the view mode for this widget
        :param mode: str
        """

        if mode == self.IconMode:
            self.set_zoom_amount(self.DEFAULT_MIN_ICON_SIZE)
        elif mode == self.TableMode:
            self.set_zoom_amount(self.DEFAULT_MIN_ICON_SIZE)

    def set_list_mode(self):
        """
        Sets the tree widget visible
        """

        self._list_view.hide()
        self._tree_widget.show()
        self._tree_widget.setFocus()

    def set_icon_mode(self):
        """
        Sets the list view visible
        """

        self._tree_widget.hide()
        self._list_view.show()
        self._list_view.setFocus()

    def zoom_amount(self):
        """
        Returns the zoom amount for the widget
        :return: int
        """

        return self._zoom_amount

    def set_zoom_amount(self, value):
        """
        Sets the zoom amount for the widget
        :param value: int
        """

        if value < self.DEFAULT_MIN_LIST_SIZE:
            value = self.DEFAULT_MIN_LIST_SIZE

        self._zoom_amount = value
        dpi = float(self.dpi())
        size = QSize(int(value * dpi), int(value * dpi))
        self.set_icon_size(size)
        if value >= self.DEFAULT_MIN_LIST_SIZE:
            self._set_view_mode(self.IconMode)
        else:
            self._set_view_mode(self.TableMode)
        column_width = value * dpi + self.item_text_height()
        self._tree_widget.setIndentation(0)
        self._tree_widget.setColumnWidth(0, column_width)
        self.scroll_to_selected_item()
        self.show_toast_message('Size: {}%'.format(value))

    def vertical_scrollbar(self):
        """
        Returns the active vertical scroll bar
        :return: QScrollBar
        """

        if self.is_table_view():
            return self.tree_widget().verticalScrollBar()
        else:
            return self.list_view().verticalScrollBar()

    def visual_item_rect(self, item):
        """
        Returns the visual rect for the item
        :param item: LibraryItem
        :return: QRect
        """

        if self.is_table_view():
            visual_rect = self.tree_widget().visual_item_rect(item)
        else:
            index = self.tree_widget().index_from_item(item)
            visual_rect = self.list_view().visualRect(index)

        return visual_rect

    def is_item_visible(self, item):
        """
        Returns whether given item is visible or not
        :param item: LibraryItem
        :return: bool
        """

        height = self.height()
        item_rect = self.visual_item_rect(item)
        scroll_bar_y = self.vertical_scrollbar().value()
        y = (scroll_bar_y - item_rect.y()) + height

        return scroll_bar_y < y < scroll_bar_y + height

    def scroll_to_item(self, item):
        """
        Ensures the given item is visible
        :param item: LibraryItem
        """

        position = QAbstractItemView.PositionAtCenter
        if self.is_table_view():
            self.tree_widget().scroll_to_item(item, position)
        elif self.is_icon_view():
            self.list_view().scroll_to_item(item, position)

    def scroll_to_selected_item(self):
        """
        Ensures that selected item is visible
        """

        item = self.selected_item()
        if item:
            self.scroll_to_item(item)

    def item_at(self, pos):
        """
        Returns the current item at the given position
        :param pos: QPoint
        :return: LibraryItem
        """

        if self.is_icon_view():
            return self.list_view().item_at(pos)
        else:
            return self.tree_widget().item_at(pos)

    def item_data(self, column_labels):
        """
        Returns all column data for the given column labels
        :param column_labels: list(str)
        :return: dict
        """

        data = dict()
        for item in self.items():
            key = item.id()
            for column_label in column_labels:
                column = self.treeWidget().column_from_label(column_label)
                value = item.data(column, Qt.EditRole)
                data.setdefault(key, dict())
                data[key].setdefault(column_label, value)

        return data

    def set_item_data(self, data):
        """
        Sets the item data for all the curren items
        :param data: dict
        """

        for item in self.items():
            key = item.id()
            if key in data:
                item.set_item_data(data[key])

    def insert_items(self, items, item_at=None):
        """
        Inserts the given items at the given position
        :param items: list(LibraryItem)
        :param item_at: LibraryItem
        """

        self.add_items(items)
        self.move_items(items, item_at=item_at)
        self.tree_widget().set_items_selected(items)

    def create_group_item(self, text, children=None):
        """
        Internal function that creates a new item for the given text and children
        :param text: str
        :param children: list(LibraryItem)
        """

        group_item = items.LibraryGroupItem()
        group_item.set_name(text)
        group_item.set_stretch_to_widget(self)
        group_item.set_children(children)

        return group_item

    def update_columns(self):
        """
        Updates the columns labels with the current item data
        """

        self.treeWidget().update_header_labels()

    def column_labels(self):
        """
        Returns all the column labels
        :return: list(str)
        """

        return self.treeWidget().column_labels()

    def refresh(self):
        """
        Refreshes the item size
        """

        self.refresh_size()

    def refresh_size(self):
        """
        Refreshes the size of the items
        """

        self.set_zoom_amount(self.zoom_amount() + 1)
        self.set_zoom_amount(self.zoom_amount() - 1)
        self.repaint()

    def toggle_text_visible(self):
        """
        Toggle the item text visibility
        """

        if self.is_item_text_visible():
            self.set_item_text_visible(False)
        else:
            self.setItemTextVisible(True)

    def is_item_text_visible(self):
        """
        Returns whether item text is visible or not
        :return: bool
        """

        if self.is_icon_view():
            return self._is_item_text_visible
        else:
            return True

    def item_text_height(self):
        """
        Returns the height of the item text
        :return: int
        """

        return self.DEFAULT_TEXT_HEIGHT * self.dpi()

    def item_delegate(self):
        """
        Returns the item delegate for the views
        :return: LibraryViewerDelegate
        """

        return self._delegate

    def set_item_text_visible(self, flag):
        """
        Sets the visibility of the item text
        :param flag: bool
        """

        self._is_item_text_visible = flag
        self.refresh_size()

    def column_labels_from_items(self):
        """
        Returns the column labels for all the items
        :return: list(str)
        """

        seq = list()
        for item in self.items():
            seq.extend(item._text_column_order)
        seen = set()

        return [x for x in seq if x not in seen and not seen.add(x)]

    def refresh_columns(self):
        """
        Refresh columns labels
        """

        self.set_column_labels(self.column_labels_from_items())

    def padding(self):
        """
        Returns the item padding
        :return: int
        """

        return self._padding

    def set_padding(self, value):
        """
        Sets the item padding
        :param value: int
        """

        if value % 2 == 0:
            self._padding = value
        else:
            self._padding = value + 1
        self.repaint()
        self.show_toast_message('Border: {}'.format(value))

    def item_size_hint(self):
        """
        Returns the item size hint
        :return: QSize
        """

        return self._item_size_hint

    def icon_size(self):
        """
        Returns the icon size for the views
        :return: QSize
        """

        return self._icon_size

    def set_icon_size(self, size):
        """
        Sets the icon size for the views
        :param size: QSize
        """

        self._icon_size = size
        if self.is_item_text_visible():
            w = size.width()
            h = size.width() + self.item_text_height()
            self._item_size_hint = QSize(w, h)
        else:
            self._item_size_hint = size

        self.list_view().setIconSize(size)
        self.tree_widget().setIconSize(size)

    def wheel_scroll_step(self):
        """
        Returns the wheel scroll setp amount
        :return: int
        """

        return self.DEFAULT_WHEEL_SCROLL_STEP

    def _set_view_mode(self, mode):
        """
        Internal function that sets the view mode ro this widget
        :param mode: str
        :return: str
        """

        if mode == self.IconMode:
            self.set_icon_mode()
        elif mode == self.TableMode:
            self.set_list_mode()

    """
    ##########################################################################################
    DPI
    ##########################################################################################
    """

    def dpi(self):
        """
        Returns zoom multiplier
        :return: int
        """

        return self._dpi

    def set_dpi(self, dpi):
        """
        Sets the zoom multiplier
        :param dpi: int
        """

        self._dpi = dpi
        self.refresh_size()

    """
    ##########################################################################################
    SETTINGS
    ##########################################################################################
    """

    def settings(self):
        """
        Returns the current state of the widget
        :return: dict
        """

        settings = dict()
        settings['columnLabels'] = self.column_labels()
        settings['padding'] = self.padding()
        settings['spacing'] = self.spacing()
        settings['zoomAmount'] = self.zoom_amount()
        settings['selectedPaths'] = self.selected_paths()
        settings['textVisible'] = self.is_item_text_visible()
        settings.update(self.tree_widget().settings())

        return settings

    def set_settings(self, settings):
        """
        Sets the current state of the widget
        :param settings: dict
        """

        self.set_toast_enabled(False)
        padding = settings.get('padding', 5)
        spacing = settings.get('spacing', 2)
        zoom_amount = settings.get('zoomAmount', 100)
        selected_paths = settings.get('selectedPaths', list())
        item_text_visible = settings.get('textVisible', True)
        self.set_padding(padding)
        self.set_spacing(spacing)
        self.set_zoom_amount(zoom_amount)
        self.select_paths(selected_paths)
        self.set_item_text_visible(item_text_visible)
        self.tree_widget().set_settings(settings)
        self.set_toast_enabled(True)

        return True

    """
    ##########################################################################################
    TOAST WIDGET
    ##########################################################################################
    """

    def toast_enabled(self):
        """
        Returns whether toast message widget is enabled or not
        :return: bool
        """

        return self._toast_enabled

    def set_toast_enabled(self, flag):
        """
        Sets whether toast widget is enabled or not
        :param flag: bool
        """

        self._toast_enabled = flag

    def show_toast_message(self, text, duration=300):
        """
        Shows a toast with the given text for the given duration
        :param text: str
        :param duration: None or int
        """

        if self.toast_enabled():
            self._toast_widget.set_duration(duration)
            self._toast_widget.setText(text)
            self._toast_widget.show()

    """
    ##########################################################################################
    TREE WIDGET
    ##########################################################################################
    """

    def tree_widget(self):
        """
        Returns the list view that contains the items
        :return: TreeWidget
        """

        return self._tree_widget

    def column_from_label(self, *args):
        """
        Returns column from given label text
        :return: int
        """

        return self.tree_widget().column_from_label(*args)

    def set_column_hidden(self, column, hidden):
        """
        Hides/Shows specific column in tree widget
        :param column: int
        :param hidden: bool
        """

        self.tree_widget().setColumnHidden(column, hidden)

    def column_labels(self):
        """
        Set all the column labels
        :return: list(str)
        """

        return self.tree_widget().column_labels()

    def set_column_labels(self, labels):
        """
        Set the columns for the viewer
        :param labels: list(str)
        """

        labels_set = set()
        set_add = labels_set.add
        labels = [x for x in labels if x.strip() and not (x in labels_set or set_add(x))]
        self.tree_widget().setHeaderLabels(labels)

    def index_from_item(self, item):
        """
        Returns the QModelIndex associated with the given item
        :param item: LibraryItem
        :return: QModelIndex
        """

        return self._tree_widget.index_from_item(item)

    def items(self):
        """
        Returns all the items in the tree widget
        :return: list(Item)
        """

        return self._tree_widget.items()

    def item_from_index(self, index):
        """
        Returns a pointer to the QTreeWidgetItem associated with the given index
        :param index: QModelIndex
        :return: QTreeWidgetItem
        """

        return self.tree_widget().itemFromIndex(index)

    def text_from_items(self, *args, **kwargs):
        """
        Returns all data for the given items and given column
        :return: list(str)
        """

        return self.tree_widget().text_from_items(*args, **kwargs)

    def text_from_column(self, *args, **kwargs):
        """
        Returns all data for the given column
        :return: list(str)
        """

        return self.tree_widget().text_from_column(*args, **kwargs)

    def add_item(self, item):
        """
        Add the item to the tree widget
        :param item: LibraryItem
        """

        self.add_items([item])

    def add_items(self, items):
        """
        Add the given items to the items widget
        :param items: list(LibraryItem)
        """

        self._tree_widget.addTopLevelItems(items)

    def selected_item(self):
        """
        Returns the last non-hidden selected item
        :return: LibraryItem
        """

        return self._tree_widget.selected_item()

    def selected_items(self):
        """
        Returns a list with all selected non-hiden items
        :return: list(QTreeWidgetItem)
        """

        return self._tree_widget.selectedItems()

    def set_item_hidden(self, item, value):
        """
        Sets the visibilty of given item
        :param item: QTreeWidgetItem
        :param value: bool
        """

        item.setHidden(value)

    def set_items_hidden(self, items, value):
        """
        Set the visibility of given items
        :param items: list(QTreeWidgetItem)
        :param value: bool
        """

        for item in items:
            self.set_item_hidden(item, value)

    def selected_paths(self):
        """
        Returns the selected item paths
        :return: list(str)
        """

        paths = list()
        for item in self.selected_items():
            path = item.url().toLocalFile()
            paths.append(path)

        return paths

    def select_paths(self, paths):
        """
        Selected the items that have the given paths
        :param paths: list(str)
        """

        for item in self.items():
            path = item.id()
            if path in paths:
                item.setSelected(True)

    def select_items(self, items):
        """
        Select the given items
        :param items: list(LibraryItem)
        """

        paths = [item.id() for item in items]
        self.select_paths(paths)

    def clear_selection(self):
        """
        Cleras the user selection
        """

        self._tree_widget.clearSelection()

    def selection_model(self):
        """
        Returns the current selection model
        :return: QItemSelectionModel
        """

        return self._tree_widget.selectionModel()

    def model(self):
        """
        Returns the model the viewer is representing
        :return: QAbstractItemModel
        """

        return self._tree_widget.model()

    def update_items(self):
        """
        Sets the items to the viewer
        """

        selected_items = self.selected_items()
        self.clear_selection()

        results = self.library().grouped_results()

        items = list()

        for group in results:
            if group != 'None':
                group_item = self.create_group_item(group)
                items.append(group_item)
            items.extend(results[group])

        self.tree_widget().set_items(items)

        if selected_items:
            self.select_items(selected_items)
            self.scroll_to_selected_item()

    def clear(self):
        """
        Clear all elements in tree widget
        """

        self.tree_widget().clear()

    def _on_update_items(self):
        self.update_items()

    """
    ##########################################################################################
    LIST WIDGET
    ##########################################################################################
    """

    def list_view(self):
        """
        Returns the list view that contains the items
        :return: LibraryListView
        """

        return self._list_view

    def set_locked(self, value):
        """
        Disables drag and drop
        :param value: bool
        """

        self.list_view().setDragEnabled(not value)
        self.list_view().setDropEnabled(not value)

    def spacing(self):
        """
        Returns the spacing between the items
        :return: int
        """

        return self.list_view().spacing()

    def set_spacing(self, spacing):
        """
        Set the spacing between the items
        :param spacing: int
        """

        self.list_view().setSpacing(spacing)
        self.scroll_to_selected_item()
        self.show_toast_message('Spacing: {}'.format(spacing))

    def move_items(self, items, item_at=None):
        """
        Moves the given items to the given position
        :param items: list(LibraryItem)
        :param item_at: LibraryItem
        """

        self.list_view().move_items(items, item_at=item_at)

    """
    ##########################################################################################
    MENUS
    ##########################################################################################
    """

    def create_item_settings_menu(self):
        """
        Cretas and returns the item settings menu for this widget
        :return: QMenu
        """

        menu = QMenu('Item View', self)

        view_settings_action = action.SeparatorAction('View Settings', menu)
        menu.addAction(view_settings_action)

        size_action = action.SliderAction('Size', menu)
        size_action.slider().setMinimum(10)
        size_action.slider().setMaximum(200)
        size_action.slider().setValue(self.zoom_amount())
        size_action.slider().valueChanged.connect(self.set_zoom_amount)
        menu.addAction(size_action)

        border_action = action.SliderAction('Border', menu)
        border_action.slider().setMinimum(0)
        border_action.slider().setMaximum(20)
        border_action.slider().setValue(self.padding())
        border_action.slider().valueChanged.connect(self.set_padding)
        menu.addAction(border_action)

        spacing_action = action.SliderAction('Spacing', menu)
        spacing_action.slider().setMinimum(self.DEFAULT_MIN_SPACING)
        spacing_action.slider().setMaximum(self.DEFAULT_MAX_SPACING)
        spacing_action.slider().setValue(self.spacing())
        spacing_action.slider().valueChanged.connect(self.set_spacing)
        menu.addAction(spacing_action)

        item_options = action.SeparatorAction('Item Options')
        menu.addAction(item_options)

        show_labels_action = QAction('Show labels', menu)
        show_labels_action.setCheckable(True)
        show_labels_action.setChecked(self.is_item_text_visible())
        show_labels_action.triggered[bool].connect(self.set_item_text_visible)
        menu.addAction(show_labels_action)

        return menu

    def create_settings_menu(self):
        """
        Creates and returns the settings menu for this widget
        :return: QMenu
        """

        menu = QMenu('Item View', self)
        menu.addSeparator()

        show_labels_action = QAction('Show labels', menu)
        show_labels_action.setCheckable(True)
        show_labels_action.setChecked(self.is_item_text_visible())
        show_labels_action.triggered[bool].connect(self.set_item_text_visible)
        menu.addAction(show_labels_action)
        menu.addSeparator()

        copy_text_menu = self.tree_widget().create_copy_text_menu()
        menu.addMenu(copy_text_menu)
        menu.addSeparator()

        size_action = action.SliderAction('Size', menu)
        size_action.slider().setMinimum(10)
        size_action.slider().setMaximum(200)
        size_action.slider().setValue(self.zoom_amount())
        size_action.slider().valueChanged.connect(self.set_zoom_amount)
        menu.addAction(size_action)

        border_action = action.SliderAction('Border', menu)
        border_action.slider().setMinimum(0)
        border_action.slider().setMaximum(20)
        border_action.slider().setValue(self.padding())
        border_action.slider().valueChanged.connect(self.set_padding)
        menu.addAction(border_action)

        spacing_action = action.SliderAction('Spacing', menu)
        spacing_action.slider().setMinimum(self.DEFAULT_MIN_SPACING)
        spacing_action.slider().setMaximum(self.DEFAULT_MAX_SPACING)
        spacing_action.slider().setValue(self.spacing())
        spacing_action.slider().valueChanged.connect(self.set_spacing)
        menu.addAction(spacing_action)

        return menu

    def create_items_menu(self, items=None):
        """
        Creates the item menu for the given items
        :param items: list(LibraryItem)
        :return: QMenu
        """

        item = items or self.selected_item()
        menu = QMenu(self)
        if item:
            try:
                item.context_menu(menu)
            except Exception as e:
                qt.logger.exception(e)
        else:
            item_action = QAction(menu)
            item_action.setText('No Item selected')
            item_action.setDisabled(True)
            menu.addAction(item_action)

        return menu

    def create_copy_text_menu(self):
        """
        Returns copy text menu for this widget
        :return: QMenu
        """

        return self.tree_widget().create_copy_text_menu()

    def create_context_menu(self):
        """
        Creates and returns the context menu for this widget
        :return: QMenu
        """

        menu = self.create_items_menu()
        settings_menu = self.create_settings_menu()
        menu.addMenu(settings_menu)

        return menu

    """
    ##########################################################################################
    PAINT
    ##########################################################################################
    """

    def text_color(self):
        """
        Returns the item text color
        :return: QColor
        """

        return self._text_color

    def set_text_color(self, color):
        """
        Sets the item text color
        :param color: QColor
        """

        self._text_color = color

    def text_selected_color(self):
        """
        Returns the item text color when selected
        :return: QColor
        """

        return self._text_selected_color

    def set_text_selected_color(self, color):
        """
        Sets the text color when an item is selected
        :param color: QColor
        """

        self._text_selected_color = color

    def background_color(self):
        """
        Returns item background color
        :return: QColor
        """

        return self._background_color

    def set_background_color(self, color):
        """
        Sets the item background color
        :param color: QColor
        """

        self._background_color = color

    def background_hover_color(self):
        """
        Returns the background color when the mouse is over an item
        :return: QColor
        """

        return self._background_hover_color

    def set_background_hover_color(self, color):
        """
        Sets the background color when the mouse hovers over the item
        :param color: QColor
        """

        self._background_hover_color = color

    def background_selected_color(self):
        """
        Returns the background color when an item is selected
        :return: QColor
        """

        return self._background_selected_color

    def set_background_selected_color(self, color):
        """
        Sets the background color when an item is selected
        :param color: QColor
        """

        self._background_selected_color = color
        self._list_view.ser_rubber_band_color(QColor(200, 200, 200, 255))

    """
    ##########################################################################################
    CALLBACKS
    ##########################################################################################
    """

    def _on_sort_indicator_changed(self):
        """
        Internal callback function that is called when sort indicator changes
        """

        pass

    def _on_item_clicked(self, item):
        """
        Internal callback function that is called when an item has been clicked
        :param item: LibraryItem
        """

        if isinstance(item, items.LibraryGroupItem):
            self.groupClicked.emit(item)
        else:
            self.itemClicked.emit(item)

    def _on_item_double_clicked(self, item):
        """
        Internal callback function that is called when an item has been double clicked
        :param item: LibraryItem
        """

        self.itemDoubleClicked.emit(item)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains extended Qt dialog classes
"""

from __future__ import print_function, division, absolute_import

import os
import string
import getpass
from functools import partial

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

import tpDcc as tp
from tpDcc import register
from tpDcc.libs.qt.core import base, qtutils, animation, theme, dragger, resizers
from tpDcc.libs.qt.widgets import layouts, dividers


class DialogContents(QFrame, object):
    """
    Widget that defines the core contents of frameless window
    Can be used to custom CSS for frameless windows contents
    """

    def __init__(self, parent=None):
        super(DialogContents, self).__init__(parent=parent)


class Dialog(QDialog, object):
    """
    Class to create basic Maya docked windows
    """

    dialogResizedFinished = Signal()
    dialogClosed = Signal()

    def __init__(self, parent=None, **kwargs):

        if parent is None:
            parent = tp.Dcc.get_main_window()

        self._setup_resizers()

        super(Dialog, self).__init__(parent=parent)

        title = kwargs.get('title', '')
        name = kwargs.get('name', title or self.__class__.__name__)
        width = kwargs.get('width', 600)
        height = kwargs.get('height', 800)
        show_on_initialize = kwargs.get('show_on_initialize', True)
        self._theme = None
        self._dpi = kwargs.get('dpi', 1.0)
        self._fixed_size = kwargs.get('fixed_size', False)
        self._has_title = kwargs.pop('has_title', False)
        self._size = kwargs.pop('size', (200, 125))
        self._title_pixmap = kwargs.pop('title_pixmap', None)

        self.setObjectName(str(name))
        self.setFocusPolicy(Qt.StrongFocus)

        frameless = kwargs.get('frameless', True)
        self.set_frameless(frameless)

        self.ui()
        self.setup_signals()

        self._status_bar = QStatusBar(self)
        self.main_layout.addWidget(self._status_bar)
        if self._fixed_size:
            self._status_bar.hide()
        self._status_bar.hide()

        self.setWindowTitle(title)

        auto_load = kwargs.get('auto_load', True)
        if auto_load:
            self.load_theme()

        if show_on_initialize:
            self.center()
            self.show()

        self.resize(width, height)

    def default_settings(self):
        """
        Returns default settings values
        :return: dict
        """

        return {
            "theme": {
                "accentColor": "rgb(80, 80, 80, 255)",
                "backgroundColor": "rgb(45, 45, 45, 255)",
            }
        }

    def load_theme(self):
        def_settings = self.default_settings()
        theme_settings = def_settings.get('theme', dict())

        self.set_theme_settings(theme_settings)

    def set_width_height(self, width, height):
        """
        Sets the width and height of the dialog
        :param width: int
        :param height: int
        """

        x = self.geometry().x()
        y = self.geometry().y()
        self.setGeometry(x, y, width, height)

    def is_frameless(self):
        """
        Returns whether or not frameless functionality for this window is enable or not
        :return: bool
        """

        return self.window().windowFlags() & Qt.FramelessWindowHint == Qt.FramelessWindowHint

    def set_frameless(self, flag):

        window = self.window()

        if flag and not self.is_frameless():
            window.setAttribute(Qt.WA_TranslucentBackground)
            if qtutils.is_pyside2() or qtutils.is_pyqt5():
                window.setWindowFlags(window.windowFlags() | Qt.FramelessWindowHint | Qt.NoDropShadowWindowHint)
            else:
                window.setWindowFlags(Qt.Window | Qt.FramelessWindowHint)
            self.set_resizer_active(True)
        elif not flag and self.is_frameless():
            window.setAttribute(Qt.WA_TranslucentBackground)
            if qtutils.is_pyside2() or qtutils.is_pyqt5():
                window.setWindowFlags(window.windowFlags() | Qt.FramelessWindowHint | Qt.NoDropShadowWindowHint)
            else:
                self.setWindowFlags(Qt.Window | Qt.FramelessWindowHint)
            self.set_resizer_active(False)

        window.show()
        #
        # if self._frameless:
        #     self.setAttribute(Qt.WA_TranslucentBackground)
        #     if qtutils.is_pyside2():
        #         self.setWindowFlags(self.windowFlags() | Qt.FramelessWindowHint)
        #     else:
        #         self.setWindowFlags(Qt.Window | Qt.FramelessWindowHint)

    def center(self, to_cursor=False):
        """
        Move the dialog to the center of the current window
        """

        frame_geo = self.frameGeometry()
        if to_cursor:
            pos = QApplication.desktop().cursor().pos()
            screen = QApplication.desktop().screenNumber(pos)
            center_point = QApplication.desktop().screenGeometry(screen).center()
        else:
            center_point = QDesktopWidget().availableGeometry().center()
        frame_geo.moveCenter(center_point)
        self.move(frame_geo.topLeft())

    def fade_close(self):
        animation.fade_window(start=1, end=0, duration=400, object=self, on_finished=self.close)

    def get_main_layout(self):
        main_layout = QVBoxLayout()
        main_layout.setContentsMargins(2, 2, 2, 2)
        main_layout.setSpacing(0)
        return main_layout

    def ui(self):

        self._central_layout = layouts.VerticalLayout(margins=(0, 0, 0, 0), spacing=0)
        self.setLayout(self._central_layout)

        self._top_widget = QWidget()
        self._top_layout = layouts.VerticalLayout(margins=(0, 0, 0, 0), spacing=0)
        self._top_widget.setLayout(self._top_layout)

        for r in self._resizers:
            r.setParent(self)
            r.windowResizedFinished.connect(self.dialogResizedFinished)
        self.set_resize_directions()

        self._dragger = dragger.DialogDragger(parent=self)
        self._dragger.setVisible(self.is_frameless())
        self._top_layout.addWidget(self._dragger)

        self.main_widget = DialogContents()
        self.main_layout = self.get_main_layout()
        self.main_widget.setLayout(self.main_layout)

        self.logo_view = QGraphicsView()
        self.logo_view.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.logo_view.setMaximumHeight(100)
        self._logo_scene = QGraphicsScene()
        self._logo_scene.setSceneRect(QRectF(0, 0, 2000, 100))
        self.logo_view.setScene(self._logo_scene)
        self.logo_view.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.logo_view.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.logo_view.setFocusPolicy(Qt.NoFocus)

        if self._has_title and self._title_pixmap:
            self._logo_scene.addPixmap(self._title_pixmap)
            self._top_layout.addWidget(self.logo_view)

        title_background_pixmap = self._get_title_pixmap()
        if self._has_title and title_background_pixmap:
            self._logo_scene.addPixmap(title_background_pixmap)
            self._top_layout.addWidget(self.logo_view)
        else:
            self.logo_view.setVisible(False)

        grid_layout = layouts.GridLayout()
        grid_layout.setHorizontalSpacing(0)
        grid_layout.setVerticalSpacing(0)
        grid_layout.setContentsMargins(0, 0, 0, 0)
        grid_layout.addWidget(self._top_widget, 1, 1, 1, 1)
        grid_layout.addWidget(self.main_widget, 2, 1, 1, 1)
        grid_layout.addWidget(self._top_left_resizer, 0, 0, 1, 1)
        grid_layout.addWidget(self._top_resizer, 0, 1, 1, 1)
        grid_layout.addWidget(self._top_right_resizer, 0, 2, 1, 1)
        grid_layout.addWidget(self._left_resizer, 1, 0, 2, 1)
        grid_layout.addWidget(self._right_resizer, 1, 2, 2, 1)
        grid_layout.addWidget(self._bottom_left_resizer, 3, 0, 1, 1)
        grid_layout.addWidget(self._bottom_resizer, 3, 1, 1, 1)
        grid_layout.addWidget(self._bottom_right_resizer, 3, 2, 1, 1)
        grid_layout.setColumnStretch(1, 1)
        grid_layout.setRowStretch(2, 1)

        # Shadow effect for dialog
        # BUG: This causes some rendering problems when using other shadow effects in child widgets of the window
        # BUG: Also detected problems when updating wigets (tree views, web browsers, etc)
        # https://bugreports.qt.io/browse/QTBUG-35196
        # shadow_effect = QGraphicsDropShadowEffect(self)
        # shadow_effect.setBlurRadius(qtutils.dpi_scale(15))
        # shadow_effect.setColor(QColor(0, 0, 0, 150))
        # shadow_effect.setOffset(qtutils.dpi_scale(0))
        # self.setGraphicsEffect(shadow_effect)

        self._central_layout.addLayout(grid_layout)

        for r in self._resizers:
            r.windowResizedFinished.connect(self.dialogResizedFinished)

        if self._size:
            self.resize(self._size[0], self._size[1])

    def statusBar(self):
        """
        Returns status bar of the dialog
        :return: QStatusBar
        """

        return self._status_bar

    def dpi(self):
        """
        Return the current dpi for the window
        :return: float
        """

        return float(self._dpi)

    def set_dpi(self, dpi):
        """
        Sets current dpi for the window
        :param dpi: float
        """

        self._dpi = dpi

    def theme(self):
        """
        Returns the current theme
        :return: Theme
        """

        if not self._theme:
            self._theme = theme.Theme()

        return self._theme

    def set_theme(self, theme):
        """
        Sets current window theme
        :param theme: Theme
        """

        self._theme = theme
        self._theme.updated.connect(self.reload_stylesheet)
        self.reload_stylesheet()

    def set_theme_settings(self, settings):
        """
        Sets the theme settings from the given settings
        :param settings: dict
        """

        # TODO: We should be able to give a dialog a theme to load
        # TODO: This will allow to setup specific themes if dialogs are launch from windows with
        # TODO: specific themes

        new_theme = tp.ResourcesMgr().theme('default')
        if not new_theme:
            new_theme = theme.Theme()
        new_theme.set_settings(settings)
        self.set_theme(new_theme)

    def reload_stylesheet(self):
        """
        Reloads the stylesheet to the current theme
        """

        current_theme = self.theme()
        current_theme.set_dpi(self.dpi())
        stylesheet = current_theme.stylesheet()
        self.setStyleSheet(stylesheet)

        all_widgets = self.main_layout.findChildren(QObject)
        for w in all_widgets:
            if hasattr(w, 'setStyleSheet'):
                w.setStyleSheet(stylesheet)
                w.update()

    def setup_signals(self):
        pass

    def set_logo(self, logo, offset=(930, 0)):
        logo = self._logo_scene.addPixmap(logo)
        logo.setOffset(offset[0], offset[1])

    def resizeEvent(self, event):
        # TODO: Take the width from the QGraphicsView not hardcoded :)
        if hasattr(self, 'logo_view'):
            self.logo_view.centerOn(1000, 0)
        return super(Dialog, self).resizeEvent(event)

    def closeEvent(self, event):
        self.dialogClosed.emit()
        event.accept()

    def setWindowIcon(self, icon):
        if self.is_frameless() or (hasattr(self, '_dragger') and self._dragger):
            self._dragger.set_icon(icon)
        super(Dialog, self).setWindowIcon(icon)

    def setWindowTitle(self, title):
        if self.is_frameless() or (hasattr(self, '_dragger') and self._dragger):
            self._dragger.set_title(title)
        super(Dialog, self).setWindowTitle(title)

    # ============================================================================================================
    # RESIZERS
    # ============================================================================================================

    def set_resizer_active(self, flag):
        """
        Sets whether resizers are enable or not
        :param flag: bool
        """

        if flag:
            for r in self._resizers:
                r.show()
        else:
            for r in self._resizers:
                r.hide()

    def set_resize_directions(self):
        """
        Sets the resize directions for the resizer widget of this window
        """

        self._top_resizer.set_resize_direction(resizers.ResizeDirection.Top)
        self._bottom_resizer.set_resize_direction(resizers.ResizeDirection.Bottom)
        self._right_resizer.set_resize_direction(resizers.ResizeDirection.Right)
        self._left_resizer.set_resize_direction(resizers.ResizeDirection.Left)
        self._top_left_resizer.set_resize_direction(resizers.ResizeDirection.Left | resizers.ResizeDirection.Top)
        self._top_right_resizer.set_resize_direction(resizers.ResizeDirection.Right | resizers.ResizeDirection.Top)
        self._bottom_left_resizer.set_resize_direction(resizers.ResizeDirection.Left | resizers.ResizeDirection.Bottom)
        self._bottom_right_resizer.set_resize_direction(
            resizers.ResizeDirection.Right | resizers.ResizeDirection.Bottom)

    def get_resizers_height(self):
        """
        Returns the total height of the vertical resizers
        :return: float
        """

        resizers = [self._top_resizer, self._bottom_resizer]
        total_height = 0
        for r in resizers:
            if not r.isHidden():
                total_height += r.minimumSize().height()

        return total_height

    def get_resizers_width(self):
        """
        Returns the total widht of the horizontal resizers
        :return: float
        """

        resizers = [self._left_resizer, self._right_resizer]
        total_width = 0
        for r in resizers:
            if not r.isHidden():
                total_width += r.minimumSize().width()

        return total_width

    def _setup_resizers(self):
        """
        Internal function that setup window resizers
        """

        self._top_resizer = resizers.VerticalResizer()
        self._bottom_resizer = resizers.VerticalResizer()
        self._right_resizer = resizers.HorizontalResizer()
        self._left_resizer = resizers.HorizontalResizer()
        self._top_left_resizer = resizers.CornerResizer()
        self._top_right_resizer = resizers.CornerResizer()
        self._bottom_left_resizer = resizers.CornerResizer()
        self._bottom_right_resizer = resizers.CornerResizer()

        self._resizers = [
            self._top_resizer, self._top_right_resizer, self._right_resizer, self._bottom_right_resizer,
            self._bottom_resizer, self._bottom_left_resizer, self._left_resizer, self._top_left_resizer
        ]

    def _get_title_pixmap(self):
        """
        Internal function that sets the pixmap used for the title
        """

        return None


class ColorDialog(Dialog, object):

    def_title = 'Select Color'

    maya_colors = [
        (.467, .467, .467), (.000, .000, .000), (.247, .247, .247), (.498, .498, .498), (0.608, 0, 0.157),
        (0, 0.016, 0.373), (0, 0, 1), (0, 0.275, 0.094), (0.145, 0, 0.263), (0.78, 0, 0.78), (0.537, 0.278, 0.2),
        (0.243, 0.133, 0.122), (0.6, 0.145, 0), (1, 0, 0), (0, 1, 0), (0, 0.255, 0.6), (1, 1, 1), (1, 1, 0),
        (0.388, 0.863, 1), (0.263, 1, 0.635), (1, 0.686, 0.686), (0.89, 0.675, 0.475), (1, 1, 0.384), (0, 0.6, 0.325),
        (0.627, 0.412, 0.188), (0.62, 0.627, 0.188), (0.408, 0.627, 0.188), (0.188, 0.627, 0.365),
        (0.188, 0.627, 0.627), (0.188, 0.404, 0.627), (0.435, 0.188, 0.627), (0.627, 0.188, 0.404)]

    def __init__(self, name='MayaColorDialog', parent=None, **kwargs):
        if parent is None:
            parent = tp.Dcc.get_main_window()

        super(ColorDialog, self).__init__(name=name, parent=parent, **kwargs)

        self._color = None

    def get_color(self):
        return self._color

    color = property(get_color)

    def ui(self):

        self.color_buttons = list()

        super(ColorDialog, self).ui()

        if tp.Dcc.get_name() == tp.Dccs.Maya and tp.Dcc.get_version() <= 2016:
            self.color_dialog = QColorDialog(parent=self)
            self.color_dialog.setWindowFlags(Qt.Widget)
            self.color_dialog.setOptions(QColorDialog.DontUseNativeDialog | QColorDialog.NoButtons)
            self.main_layout.addWidget(self.color_dialog)
        else:
            grid_layout = QGridLayout()
            grid_layout.setAlignment(Qt.AlignTop)
            self.main_layout.addLayout(grid_layout)
            color_index = 0
            for i in range(0, 4):
                for j in range(0, 8):
                    color_btn = QPushButton()
                    color_btn.setMinimumHeight(35)
                    color_btn.setMinimumWidth(35)
                    self.color_buttons.append(color_btn)
                    color_btn.setStyleSheet('background-color:rgb(%s,%s,%s);' % (
                        self.maya_colors[color_index][0] * 255,
                        self.maya_colors[color_index][1] * 255,
                        self.maya_colors[color_index][2] * 255
                    ))
                    grid_layout.addWidget(color_btn, i, j)
                    color_index += 1
            selected_color_layout = QHBoxLayout()
            self.main_layout.addLayout(selected_color_layout)
            self.color_slider = QSlider(Qt.Horizontal)
            self.color_slider.setMinimum(0)
            self.color_slider.setMaximum(31)
            self.color_slider.setValue(2)
            self.color_slider.setStyleSheet(
                "QSlider::groove:horizontal {border: 1px solid #999999;height: 25px; /* the groove expands "
                "to the size of the slider by default. by giving it a height, it has a fixed size */background: "
                "qlineargradient(x1:0, y1:0, x2:0, y2:1, stop:0 #B1B1B1, stop:1 #c4c4c4);margin: 2px 0;}"
                "QSlider::handle:horizontal {background: qlineargradient(x1:0, y1:0, x2:1, y2:1, stop:0 #b4b4b4,"
                " stop:1 #8f8f8f);border: 1px solid #5c5c5c;width: 10px;margin: -2px 0; /* handle is placed by "
                "default on the contents rect of the groove. Expand outside the groove */border-radius: 1px;}")
            selected_color_layout.addWidget(self.color_slider)

            color_label_layout = QHBoxLayout()
            color_label_layout.setContentsMargins(10, 10, 10, 0)
            self.main_layout.addLayout(color_label_layout)

            self.color_lbl = QLabel()
            self.color_lbl.setStyleSheet("border: 1px solid black; background-color:rgb(0, 0, 0);")
            self.color_lbl.setMinimumWidth(45)
            self.color_lbl.setMaximumWidth(80)
            self.color_lbl.setMinimumHeight(80)
            self.color_lbl.setAlignment(Qt.AlignCenter)
            color_label_layout.addWidget(self.color_lbl)

        bottom_layout = QHBoxLayout()
        bottom_layout.setAlignment(Qt.AlignRight)
        self.main_layout.addLayout(bottom_layout)

        self.ok_btn = QPushButton('Ok')
        self.cancel_btn = QPushButton('Cancel')
        bottom_layout.addLayout(dividers.DividerLayout())
        bottom_layout.addWidget(self.ok_btn)
        bottom_layout.addWidget(self.cancel_btn)

    def setup_signals(self):

        if tp.Dcc.get_name() == tp.Dccs.Maya and tp.Dcc.get_version() <= 2016:
            pass
        else:
            for i, btn in enumerate(self.color_buttons):
                btn.clicked.connect(partial(self._on_set_color, i))
            self.color_slider.valueChanged.connect(self._on_set_color)

        self.ok_btn.clicked.connect(self._on_ok_btn)
        self.cancel_btn.clicked.connect(self._on_cancel_btn)

    def _on_set_color(self, color_index):

        if tp.Dcc.get_name() == tp.Dccs.Maya and tp.Dcc.get_version() <= 2016:
            self.color_dialog.setCurrentColor(QColor.fromRgb(
                self.maya_colors[color_index][0] * 255,
                self.maya_colors[color_index][1] * 255,
                self.maya_colors[color_index][2] * 255
            ))
        else:
            self.color_lbl.setStyleSheet('background-color:rgb(%s,%s,%s);' % (
                self.maya_colors[color_index][0] * 255,
                self.maya_colors[color_index][1] * 255,
                self.maya_colors[color_index][2] * 255
            ))
            self.color_slider.setValue(color_index)

    def _on_set_slider(self, color_index):
        self._set_color(color_index=color_index)

    def _on_ok_btn(self):

        if tp.Dcc.get_name() == tp.Dccs.Maya and tp.Dcc.get_version() <= 2016:
            pass
        else:
            self._color = self.color_slider.value()

        self.close()

    def _on_cancel_btn(self):
        self._color = None
        self.close()


class BaseFileFolderDialog(Dialog, object):
    """
    Base dialog classes for folders and files
    """

    def_title = 'Select File'
    def_size = (200, 125)
    def_use_app_browser = False

    def __init__(self,
                 name='BaseFileFolder', parent=None, **kwargs):
        super(BaseFileFolderDialog, self).__init__(name=name, parent=parent)

        self.directory = None
        self.filters = None
        self._use_app_browser = kwargs.pop('use_app_browser', self.def_use_app_browser)

        self.set_filters('All Files (*.*)')

        # By default, we set the directory to the user folder
        self.set_directory(os.path.expanduser('~'))
        self.center()

    def open_app_browser(self):
        return

    def ui(self):
        super(BaseFileFolderDialog, self).ui()

        from tpDcc.libs.qt.widgets import directory

        self.places = dict()

        self.grid = QGridLayout()
        sub_grid = QGridLayout()
        self.grid.addWidget(QLabel('Path:'), 0, 0, Qt.AlignRight)

        self.path_edit = QLineEdit(self)
        self.path_edit.setReadOnly(True)
        self.filter_box = QComboBox(self)
        self.file_edit = QLineEdit(self)

        self.view = directory.FileListWidget(self)
        self.view.setWrapping(True)
        self.view.setFocusPolicy(Qt.StrongFocus)

        self.open_button = QPushButton('Select', self)
        self.cancel_button = QPushButton('Cancel', self)

        size = QSize(32, 24)
        self.up_button = QPushButton('Up')
        self.up_button.setToolTip('Go up')
        self.up_button.setMinimumSize(size)
        self.up_button.setMaximumSize(size)

        size = QSize(56, 24)
        self.refresh_button = QPushButton('Reload')
        self.refresh_button.setToolTip('Reload file list')
        self.refresh_button.setMinimumSize(size)
        self.refresh_button.setMaximumSize(size)

        self.show_hidden = QCheckBox('Hidden')
        self.show_hidden.setChecked(False)
        self.show_hidden.setToolTip('Toggle show hidden files')

        sub_grid.addWidget(self.up_button, 0, 1)
        sub_grid.addWidget(self.path_edit, 0, 2)
        sub_grid.addWidget(self.refresh_button, 0, 3)
        sub_grid.addWidget(self.show_hidden, 0, 4)
        self.grid.addLayout(sub_grid, 0, 1)
        self.grid.addWidget(self.get_drives_widget(), 1, 0)
        self.grid.addWidget(self.view, 1, 1)
        self.grid.addWidget(QLabel('File name:'), 7, 0, Qt.AlignRight)
        self.grid.addWidget(self.file_edit, 7, 1)
        self.filter_label = QLabel('Filter:')
        self.grid.addWidget(self.filter_label, 8, 0, Qt.AlignRight)
        self.grid.addWidget(self.filter_box, 8, 1)
        hbox = QGridLayout()
        hbox.addWidget(self.open_button, 0, 0, Qt.AlignRight)
        hbox.addWidget(self.cancel_button, 0, 1, Qt.AlignRight)
        self.grid.addLayout(hbox, 9, 1, Qt.AlignRight)
        self.main_layout.addLayout(self.grid)
        self.setGeometry(200, 100, 600, 400)

        self.open_button.clicked.connect(self.accept)
        self.cancel_button.clicked.connect(self.reject)
        self.up_button.clicked.connect(self.go_up)
        self.refresh_button.clicked.connect(self.update_view)
        self.show_hidden.stateChanged.connect(self.update_view)
        self.view.directory_activated.connect(self.activate_directory_from_view)
        self.view.file_activated.connect(self.activate_file_from_view)
        self.view.file_selected.connect(self.select_file_item)
        self.view.folder_selected.connect(self.select_folder_item)
        self.view.up_requested.connect(self.go_up)
        self.view.update_requested.connect(self.update_view)

    def exec_(self, *args, **kwargs):
        if self._use_app_browser:
            return self.open_app_browser()
        else:
            self.update_view()
            self.filter_box.currentIndexChanged.connect(self.update_view)
            accepted = super(BaseFileFolderDialog, self).exec_()
            self.filter_box.currentIndexChanged.disconnect(self.update_view)
            return self.get_result() if accepted == 1 else None

    def set_filters(self, filters, selected=0):
        self.filter_box.clear()
        filter_types = filters.split(';;')
        for ft in filter_types:
            extensions = string.extract(ft, '(', ')')
            filter_name = string.rstrips(ft, '({})'.format(extensions))
            extensions = extensions.split(' ')
            self.filter_box.addItem('{} ({})'.format(filter_name, ','.join(extensions)), extensions)
        if 0 <= selected < self.filter_box.count():
            self.filter_box.setCurrentIndex(selected)
        self.filters = filters

    def get_drives_widget(self):
        """
        Returns a QGroupBox widget that contains all disk drivers of the PC in a vertical layout
        :return: QGroupBox
        """

        w = QGroupBox('')
        w.setParent(self)
        box = QVBoxLayout()
        box.setAlignment(Qt.AlignTop)
        places = [(getpass.getuser(), os.path.realpath(os.path.expanduser('~')))]
        places += [(q, q) for q in [os.path.realpath(x.absolutePath()) for x in QDir().drives()]]
        for label, loc in places:
            icon = QFileIconProvider().icon(QFileInfo(loc))
            drive_btn = QRadioButton(label)
            drive_btn.setIcon(icon)
            drive_btn.setToolTip(loc)
            drive_btn.setProperty('path', loc)
            drive_btn.clicked.connect(self.go_to_drive)
            self.places[loc] = drive_btn
            box.addWidget(drive_btn)
        w.setLayout(box)
        return w

    def go_to_drive(self):
        """
        Updates widget to show the content of the selected disk drive
        """

        sender = self.sender()
        self.set_directory(sender.property('path'), False)

    def get_result(self):
        tf = self.file_edit.text()
        sf = self.get_file_path(tf)
        return sf, os.path.dirname(sf), tf.split(os.pathsep)

    def get_filter_patterns(self):
        """
        Get list of filter patterns that are being used by the widget
        :return: list<str>
        """

        idx = self.filter_box.currentIndex()
        if idx >= 0:
            return self.filter_box.itemData(idx)
        else:
            return []

    def get_file_path(self, file_name):
        """
        Returns file path of the given file name taking account the selected directory
        :param file_name: str, name of the file without path
        :return: str
        """

        sname = file_name.split(os.pathsep)[0]
        return os.path.realpath(os.path.join(os.path.abspath(self.directory), sname))

#     def accept(self):
#         self._overlay.close()
#         super(BaseFileFolderDialog, self).accept()
#
#
#     def reject(self):
#         self._overlay.close()
#         super(BaseFileFolderDialog, self).reject()

    def update_view(self):
        """
        Updates file/folder view
        :return:
        """

        self.view.clear()
        qdir = QDir(self.directory)
        qdir.setNameFilters(self.get_filter_patterns())
        filters = QDir.Dirs | QDir.AllDirs | QDir.Files | QDir.NoDot | QDir.NoDotDot
        if self.show_hidden.isChecked():
            filters = filters | QDir.Hidden
        entries = qdir.entryInfoList(filters=filters, sort=QDir.DirsFirst | QDir.Name)
        file_path = self.get_file_path('..')
        if os.path.exists(file_path) and file_path != self.directory:
            icon = QFileIconProvider().icon(QFileInfo(self.directory))
            QListWidgetItem(icon, '..', self.view, 0)
        for info in entries:
            icon = QFileIconProvider().icon(info)
            suf = info.completeSuffix()
            name, tp = (
                info.fileName(), 0) if info.isDir() else ('%s%s' % (info.baseName(), '.%s' % suf if suf else ''), 1)
            QListWidgetItem(icon, name, self.view, tp)
        self.view.setFocus()

    def set_directory(self, path, check_drive=True):
        """
        Sets the directory that you want to explore
        :param path: str, valid path
        :param check_drive: bool,
        :return:
        """

        self.directory = os.path.realpath(path)
        self.path_edit.setText(self.directory)
        self.file_edit.setText('')

        # If necessary, update selected disk driver
        if check_drive:
            for loc in self.places:
                rb = self.places[loc]
                rb.setAutoExclusive(False)
                rb.setChecked(loc.lower() == self.directory.lower())
                rb.setAutoExclusive(True)

        self.update_view()
        self.up_button.setEnabled(not self.cant_go_up())

    def go_up(self):
        """
        Updates the current directory to go to its parent directory
        """

        self.set_directory(os.path.dirname(self.directory))

    def cant_go_up(self):
        """
        Checks whether we can naviage to current selected parent directory or not
        :return: bool
        """

        return os.path.dirname(self.directory) == self.directory

    def activate_directory_from_view(self, name):
        """
        Updates selected directory
        :param name: str, name of the directory
        """

        self.set_directory(os.path.join(self.directory, name))

    def activate_file_from_view(self, name):
        """
        Updates selected file text and returns its info by accepting it
        :param name: str, name of the file
        """

        self.select_file_item(name=name)
        self.accept()

    def select_file_item(self, name):
        """
        Updates selected file text and returns its info by accepting it
        :param name: str, name of the file
        """

        self.file_edit.setText(name)

    def select_folder_item(self, name):
        """
        Updates selected folder text and returns its info by accepting it
        :param name: str, name of the folder
        """

        self.file_edit.setText(name)


class OpenFileDialog(BaseFileFolderDialog, object):
    """
    Open file dialog
    """

    def __init__(
            self,
            name='OpenFile',
            multi=False,
            title='Open File',
            size=(200, 125),
            fixed_size=False,
            frame_less=True,
            hide_title=False,
            parent=None,
            use_app_browser=False):

        if parent is None:
            parent = tp.Dcc.get_main_window()

        super(OpenFileDialog, self).__init__(
            name=name, title=title, size=size, fixed_size=fixed_size, frame_less=frame_less,
            hide_title=hide_title, use_app_browser=use_app_browser, parent=parent
        )

        self._multi = multi
        if multi:
            self.setExtendedSelection()

    def accept(self, *args, **kwargs):
        selected_file, selected_dir, selected_file_name = self.get_result()
        if not os.path.isdir(selected_file):
            if os.path.exists(selected_file):
                super(OpenFileDialog, self).accept()
            else:
                message_box = QMessageBox()
                message_box.setWindowTitle('Confirme file selection')
                message_box.setText('File "{0}" does not exists!'.format(selected_file))
                message_box.exec_()

    def open_app_browser(self):
        if tp.Dcc.get_name() == tp.Dccs.Maya:
            import maya.cmds as cmds
            sel_file = cmds.fileDialog2(
                caption=self.windowTitle(),
                fileMode=1,
                fileFilter=self.filters,
                dialogStyle=2
            )
        else:
            raise NotImplementedError(
                'Open App Browser is not implemented for your current DCC: {}'.format(tp.Dcc.get_name()))

        if sel_file:
            sel_file = sel_file[0]
            return [sel_file, os.path.dirname(sel_file), [os.path.basename(sel_file)]]

        return None

    def select_file_item(self, names):
        if self._multi:
            self.file_edit.setText(os.pathsep.join(names))
        else:
            super(OpenFileDialog, self).select_file_item(names)


class SaveFileDialog(BaseFileFolderDialog, object):
    def __init__(self,
                 name='SaveFile',
                 title='Save File',
                 size=(200, 125),
                 fixed_size=False,
                 frame_less=True,
                 hide_title=False,
                 parent=None,
                 use_app_browser=False):

        if parent is None:
            parent = tp.Dcc.get_main_window()

        super(SaveFileDialog, self).__init__(
            name=name, title=title, size=size, fixed_size=fixed_size, frame_less=frame_less, hide_title=hide_title,
            use_app_browser=use_app_browser, parent=parent)

        self._open_button.setText('Save')
        size = QSize(42, 24)
        self.new_directory_button = QPushButton('New')
        self.new_directory_button.setToolTip('Create new directory')
        self.new_directory_button.setMinimumSize(size)
        self.new_directory_button.setMaximumWidth(size)
        self.new_directory_button.clicked.connect(self.create_new_directory)
        self.grid.itemAtPosition(0, 1).addWidget(self.new_directory_button, 0, 5)

    def accept(self, *args, **kwargs):
        selected_file, selected_dir, selected_filename = self.get_result()
        if not os.path.isdir(selected_file):
            if os.path.exists(selected_file):
                message_box = QMessageBox()
                message_box.setWindowTitle('Confirm File Selection')
                message_box.setText('File "%s" exists.\nDo you want to overwrite it?' % selected_file)
                message_box.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
                message_box.setDefaultButton(QMessageBox.No)
                rv = message_box.exec_()
                if rv == QMessageBox.Yes and not os.path.isdir(selected_file):
                    super(SaveFileDialog, self).accept()
        else:
            super(SaveFileDialog, self).accept()

    def open_app_browser(self):

        if tp.Dcc.get_name() == tp.Dccs.Maya:
            import maya.cmds as cmds
            sel_file = cmds.fileDialog2(
                caption=self.windowTitle(),
                fileMode=0,
                fileFilter=self.filters,
                dialogStyle=2
            )
        else:
            raise NotImplementedError(
                'Open App Browser is not implemented for your current DCC: {}'.format(tp.Dcc.get_name()))

        if sel_file:
            sel_file = sel_file[0]

        return [sel_file, os.path.dirname(sel_file), [os.path.basename(sel_file)]]

    def create_new_directory(self):
        name, ok = QInputDialog.getText(self, 'New directory name', 'Name:', QLineEdit.Normal, 'New Directory')
        if ok and name:
            path = os.path.join(self.directory, name)
            if os.path.exists(path):
                msg_box = QMessageBox(self)
                msg_box.setWindowTitle('Error')
                msg_box.setText('Directory already exists')
                msg_box.exec_()
            else:
                try:
                    os.makedirs(path)
                    self.update_view()
                except os.error as e:
                    msg_box = QMessageBox(self)
                    msg_box.setWindowTitle('Error')
                    msg_box.setText('Cannot create directory')
                    msg_box.exec_()


class SelectFolderDialog(BaseFileFolderDialog, object):
    def __init__(self,
                 name='SelectFolder',
                 title='Select Folder',
                 size=(200, 125),
                 fixed_size=False,
                 frame_less=True,
                 hide_title=False,
                 use_app_browser=False,
                 parent=None,
                 **kwargs):

        if parent is None:
            parent = tp.Dcc.get_main_window()

        super(SelectFolderDialog, self).__init__(
            name=name, title=title, size=size, fixed_size=fixed_size, frame_less=frame_less, hide_title=hide_title,
            use_app_browser=use_app_browser, parent=parent, **kwargs
        )

    def accept(self, *args, **kwargs):
        selected_file, selected_dir, selected_filename = self.get_result()
        super(SelectFolderDialog, self).accept()

    def open_app_browser(self):

        if tp.Dcc.get_name() == tp.Dccs.Maya:
            import maya.cmds as cmds
            sel_folder = cmds.fileDialog2(
                caption=self.windowTitle(),
                fileMode=3,
                fileFilter=self.filters,
                dialogStyle=2
            )
        else:
            raise NotImplementedError(
                'Open App Browser is not implemented for your current DCC: {}'.format(tp.Dcc.get_name()))

        if sel_folder:
            sel_folder = sel_folder[0]

            result = [sel_folder, os.path.dirname(sel_folder), [os.path.basename(sel_folder)]]
            return result[0]

        return None

    def exec_(self, *args, **kwargs):
        self.set_filters('')
        return super(SelectFolderDialog, self).exec_()


class NativeDialog(object):
    """
    Dialog that opens DCC native dialogs
    """

    @staticmethod
    def open_file(title='Open File', start_directory=None, filters=None):
        """
        Function that shows open file DCC native dialog
        :param title: str
        :param start_directory: str
        :param filters: str
        :return: str
        """

        raise NotImplementedError('open_file() function is not implemented')

    @staticmethod
    def save_file(title='Save File', start_directory=None, filters=None):
        """
        Function that shows save file DCC native dialog
        :param title: str
        :param start_directory: str
        :param filters: str
        :return: str
        """

        raise NotImplementedError('save_file() function is not implemented')

    @staticmethod
    def select_folder(title='Select Folder', start_directory=None):
        """
        Function that shows select folder DCC native dialog
        :param title: str
        :param start_directory: str
        :return: str
        """

        raise NotImplementedError('select_folder() function is not implemented')


register.register_class('Dialog', Dialog)
register.register_class('OpenFileDialog', OpenFileDialog)
register.register_class('SaveFileDialog', SaveFileDialog)
register.register_class('SelectFolderDialog', SelectFolderDialog)
register.register_class('NativeDialog', NativeDialog)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains classes that extend QToolBar functionality
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *

import tpDcc
from tpDcc.libs.qt.core import base, icon, qtutils
from tpDcc.libs.qt.widgets import layouts, buttons


class ToolBar(QToolBar, object):
    """
    Class that adds functionality to expand/collapse QToolBars
    """

    DEFAULT_EXPANDED_HEIGHT = 34
    DEFAULT_COLLAPSED_HEIGHT = 10

    def __init__(self, *args):
        super(ToolBar, self).__init__(*args)

        self._is_expanded = True
        self._expanded_height = self.DEFAULT_EXPANDED_HEIGHT
        self._collapsed_height = self.DEFAULT_COLLAPSED_HEIGHT

    def mousePressEvent(self, event):
        if not self.is_expanded():
            self.expand()

    def setFixedHeight(self, value):
        """
        Overrides base QToolBar setFixedHeight
        Allows to also set the height for all child widgets of the menu bar
        :param value: float
        """

        self.set_children_height(value)
        super(QToolBar, self).setFixedHeight(value)

    def insertAction(self, before, action):
        """
        Overrides base QToolBar insertAction function
        Support the before argument as string
        :param before: QAction or str
        :param action: QAction
        :return: QAction
        """

        action.setParent(self)
        if isinstance(before, (unicode, str)):
            before = self.find_action(before)

        action = super(ToolBar, self).insertAction(before, action)

        return action

    def actions(self):
        """
        Overrides base QToolBar actions function
        Returns all the widgets that are a child of the menu bar widget
        :return: list(QWidget)
        """

        widgets = list()
        for i in range(self.layout().count()):
            w = self.layout().itemAt(i).widget()
            if isinstance(w, QWidget):
                widgets.append(w)

        return widgets

    def widgets(self):
        """
        Returns all the widgets that are a child of the menu bar widget
        :return: list(QWidget)
        """

        widgets = list()
        for i in range(self.layout().count()):
            w = self.layout().itemAt(i).widget()
            if isinstance(w, QWidget):
                widgets.append(w)

        return widgets

    def is_expanded(self):
        """
        Returns whether the menu bar is expanded or not
        :return: bool
        """

        return self._is_expanded

    def expand_height(self):
        """
        Returns the height of menu bar when is expanded
        :return: float
        """

        return self._expanded_height

    def collapse_height(self):
        """
        Returns the height of widget when collapsed
        :return: int
        """

        return self._collapsed_height

    def set_children_hidden(self, flag):
        """
        Hide/Show all child widgets
        :param flag: bool
        """

        for w in self.widgets():
            w.setHidden(flag)

    def set_children_height(self, height):
        """
        Set the height of all the child widgets to the given height
        :param height: int
        """

        for w in self.widgets():
            w.setFixedHeight(height)

    def expand(self):
        """
        Expand the menu bar to the expand height
        """

        self._is_expanded = True
        height = self.expand_height()
        self.setFixedHeight(height)
        self.set_children_hidden(False)
        self.setIconSize(QSize(height, height))

    def collapse(self):
        """
        Collapse the menu bar to the collapse height
        """

        self._is_expanded = False
        height = self.collapse_height()
        self.setFixedHeight(height)
        self.set_children_height(0)
        self.set_children_hidden(True)
        self.setIconSize(QSize(0, 0))

    def set_icon_color(self, color):
        """
        Set the icon colors to the current foregroundRole
        :param color: QColor
        """

        for action in self.actions():
            action_icon = action.icon()
            action_icon = icon.Icon(action_icon)
            action_icon.set_color(color)
            action.setIcon(action_icon)

    def find_action(self, text):
        """
        Find the action with the given text
        :param text: str
        :return: QAction or None
        """

        for child in self.children():
            if isinstance(child, QAction):
                if child.text() == text:
                    return child

    def find_tool_button(self, text):
        """
        Find the QToolButton with the given text
        :param text: str
        :return: QToolButton or None
        """

        for child in self.children():
            if isinstance(child, QAction):
                if child.text() == text:
                    return self.widgetForAction(child)


class FlowToolBar(QFrame, object):
    """
    Custom toolbar whose buttons will flow from left to right and wrap to next row if there is no space
    """

    def __init__(self, menu_indicator_icon=None, icon_size=20, icon_padding=2, parent=None):
        super(FlowToolBar, self).__init__(parent)

        self._icon_size = icon_size
        self._icon_padding = icon_padding
        self._overflow_button_color = (128, 128, 128)
        self._menu_indicator_icon = menu_indicator_icon or tpDcc.ResourcesMgr().icon('arrow_menu')
        self._overflow_icon = tpDcc.ResourcesMgr().icon('sort_down')
        self._overflow_menu = False
        self._overflow_menu_button = None

        self.ui()

    @property
    def flow_layout(self):
        return self._flow_layout

    @property
    def overflow_layout(self):
        return self._overflow_layout

    def resizeEvent(self, event):
        self.update_widgets_overflow(event.size())

    def sizeHint(self):
        spacing_x = self._flow_layout.spacing_x
        next_x = 0
        for item in self._flow_layout.items_list:
            widget = item.widget()
            next_x += widget.sizeHint().width() + spacing_x

        return QSize(next_x + 3, super(FlowToolBar, self).sizeHint().height())

    def ui(self):
        main_layout = layouts.HorizontalLayout(margins=(0, 0, 0, 0), spacing=0)
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setAlignment(Qt.AlignTop)
        self.setLayout(main_layout)

        self._flow_layout = layouts.FlowLayout(margin=0, spacing_x=1, spacing_y=1)
        main_layout.addLayout(self._flow_layout)

        self._overflow_menu_button = self._setup_overflow_menu_button()
        self._overflow_menu = FlowToolbarMenu(parent=self)
        self._overflow_layout = self._overflow_menu.layout()

    def get_icon_size(self):
        """
        Returns the icon size
        :return: QSize
        """

        return QSize(self._icon_size + self._icon_padding, self._icon_size + self._icon_padding)

    def set_icon_size(self, size):
        """
        Sets the size of the icons of the toolbar
        :param size: QSize
        """

        self._icon_size = size

        for i in range(0, self._flow_layout.count()):
            widget = self._flow_layout.itemAt(i).widget()
            widget.setIconSize(self.get_icon_size())

        self._overflow_menu_button = self._setup_overflow_menu_button(btn=self._overflow_menu_button)

    def set_icon_padding(self, padding):
        """
        Sets the padding for the icons of the toolbar
        :param padding: int
        """

        self._icon_padding = qtutils.dpi_scale(padding)

    def set_overflow_button_color(self, button_color):
        """
        Sets the color used for the overflow button
        :param button_color:
        """

        self._overflow_button_color = button_color

    def set_height(self, height):
        """
        Sets fixed height for the toolbar
        :param height: int
        """

        self.setFixedHeight(height)

    def set_spacing_x(self, value):
        """
        Sets spacing of items in layout in X
        :param value: float
        """

        self._flow_layout.set_spacing_x(value)

    def set_spacing_y(self, value):
        """
        Sets spacing of items in layout in Y
        :param value: float
        """

        self._flow_layout.set_spacing_y(value)

    def items_list(self):
        """
        Returns list of item in toolbar without the overflow menu button
        :return: list
        """

        return self._flow_layout.items_list[:-1]

    def items(self):
        """
        Returns all items in the toolbar
        :return: list
        """

        return self._flow_layout.items()

    def update_widgets_overflow(self, size=None):
        """
        Function that hides or show widgets based on the size of the flow toolbar
        If it is too small, it will move widgets to overflow menu and if there are widget in the overflow menu,
        placce it back into the flow toolbar if there is space
        :param size: QSize, new size
        """

        if not self._overflow_menu_button or not self._overflow_menu:
            return

        spacing_x = self._flow_layout.spacing_x
        spacing_y = self._flow_layout.spacing_y

        if not size:
            size = self.size()
        if len(self.items_list()) == 0:
            return

        overflow_button_width = self._overflow_menu_button.sizeHint().width()
        width = size.width() - overflow_button_width - spacing_x
        height = size.height()
        hidden = list()

        self.setUpdatesEnabled(False)

        next_x = 0
        next_y = self.items_list()[0].widget().height()

        for item in self.items_list():
            item_widget = item.widget()
            widget_width = item_widget.sizeHint().width() + spacing_x
            next_x += widget_width
            if next_x > width:
                next_y += item_widget.height() + (spacing_y * 2)
                next_x = 0
            if next_y > height:
                item_widget.hide()
                hidden.append(item_widget)
            else:
                item_widget.show()

        menu = self._overflow_menu_button.menu(mouse_menu=Qt.LeftButton)
        for a in menu.actions():
            a.setVisible(False)

        for hidden_widget in hidden:
            for a in menu.actions():
                if a.text() == hidden_widget.property('name'):
                    a.setVisible(True)
                    break

        self._overflow_menu_button.setVisible(len(hidden) > 0)
        self.setUpdatesEnabled(True)

        return hidden

    def clear(self):
        """
        Clears all toolbar widgets
        """

        self._overflow_menu_button.clear_menu(Qt.LeftButton)
        self._flow_layout.removeWidget(self._overflow_menu_button)
        self._flow_layout.clear()
        qtutils.clear_layout(self._overflow_layout)

    def overflow_menu_active(self, flag):
        self._overflow_menu = flag
        self._overflow_menu_button.setVisible(flag)

    def _setup_overflow_menu_button(self, btn=None):
        """
        Internal function that setup overflow menu and connects it to given button.
        If button is not given, it will be created
        :param btn:
        :return:
        """

        overflow_color = self._overflow_button_color
        overflow_icon = self._overflow_icon
        if not btn:
            btn = buttons.IconMenuButton(parent=self)
        btn.set_icon(overflow_icon, colors=overflow_color, size=self._icon_size, color_offset=40)
        btn.double_click_enabled = False
        btn.setProperty('name', 'overflow')
        btn.setIconSize(self.get_icon_size())
        btn.setVisible(False)

        return btn


class FlowToolbarMenu(tpDcc.Dialog, object):
    def __init__(self, parent=None):
        super(FlowToolbarMenu, self).__init__(parent=parent, show_on_initialize=False)

    def ui(self):
        super(FlowToolbarMenu, self).ui()

        self.setWindowFlags(Qt.FramelessWindowHint | Qt.Popup)

    def sizeHint(self):
        return self.minimumSize()

    def show(self, *args, **kwargs):
        super(FlowToolbarMenu, self).show(*args, **kwargs)
        self.resize(self.sizeHint())

    def layout(self):
        return self.main_layout



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains custom widget for Qt check boxes
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

from tpDcc.libs.qt.core import mixin, animation


@mixin.cursor_mixin
class BaseCheckBox(QCheckBox, object):
    def __init__(self, text='', parent=None):
        super(BaseCheckBox, self).__init__(text=text, parent=parent)


class StyledBaseCheckBox(BaseCheckBox, animation.BaseAnimObject):
    _glow_brushes = dict()
    for index in range(1, 11):
        _glow_brushes[index] = [QBrush(QColor(0, 255, 0, 1 * index)),
                                QBrush(QColor(0, 255, 0, 3 * index)),
                                QBrush(QColor(0, 255, 0, 15 * index)),
                                QBrush(QColor(0, 255, 0, 25.5 * index))]

    _disabled_glow_brushes = {}
    for index in range(1, 11):
        _disabled_glow_brushes[index] = [QBrush(QColor(125, 125, 125, 1 * index)),
                                         QBrush(QColor(125, 125, 125, 3 * index)),
                                         QBrush(QColor(125, 125, 125, 15 * index)),
                                         QBrush(QColor(125, 125, 125, 25.5 * index))]

    def __init__(self, *args, **kwargs):
        QCheckBox.__init__(self, *args, **kwargs)
        animation.BaseAnimObject.__init__(self)

    def paintEvent(self, event):
        painter = QStylePainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        painter.setRenderHint(QPainter.TextAntialiasing)
        option = QStyleOption()
        option.initFrom(self)

        x = option.rect.x()
        y = option.rect.y()
        height = option.rect.height() - 1
        width = option.rect.width() - 1
        font = self.font()
        text = self.text()
        alignment = (Qt.AlignLeft | Qt.AlignVCenter)

        painter.setPen(self._pens_border)
        painter.setBrush(self._brush_border)
        painter.drawRoundedRect(QRect(x + 2, y + 2, 13, 13), 3, 3)

        if self.isEnabled():
            painter.setPen(self._pens_shadow)
            painter.drawText(21, y + 2, width, height, alignment, text)

            painter.setPen(self._pens_text)
            painter.drawText(20, y + 1, width, height, alignment, text)

        else:
            painter.setPen(self._pens_shadow_disabled)
            painter.drawText(21, y + 2, width, height, alignment, text)

            painter.setPen(self._pens_text_disabled)
            painter.drawText(20, y + 1, width, height, alignment, text)

        painter.setPen(self._pens_clear)

        if self.isEnabled():
            glow_brushes = self._glow_brushes
        else:
            glow_brushes = self._disabled_glow_brushes

        if self.checkState():
            for index, pos, size, corner in zip(range(4), (2, 3, 4, 5), (13, 11, 9, 7), (4, 3, 3, 2)):
                painter.setBrush(glow_brushes[10][index])
                painter.drawRoundedRect(QRect(x + pos, y + pos, size, size), corner, corner)

        glow_index = self._glow_index
        if glow_index > 0:
            for index, pos, size, corner in zip(range(4), (3, 4, 5, 6), (11, 9, 7, 5), (3, 3, 2, 2)):
                painter.setBrush(glow_brushes[glow_index][index])
                painter.drawRoundedRect(QRect(x + pos, y + pos, size, size), corner, corner)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains base functionality for Qt widgets
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *

from tpDcc.libs.qt.core import qtutils, mixin, theme


@mixin.theme_mixin
@mixin.cursor_mixin
class BaseWidget(QWidget, object):
    """
    Base class for all QWidgets based items
    """

    def_use_scrollbar = False

    def __init__(self, parent=None, **kwargs):
        super(BaseWidget, self).__init__(parent=parent)

        self._size = self.theme_default_size()

        self._use_scrollbar = kwargs.get('use_scrollbar', self.def_use_scrollbar)

        self.ui()
        self.setup_signals()

    # =================================================================================================================
    # PROPERTIES
    # =================================================================================================================

    def _get_size(self):
        """
        Returns the spin box height size
        :return: float
        """

        return self._size

    def _set_size(self, value):
        """
        Sets spin box height size
        :param value: float
        """

        self._size = value
        self.style().polish(self)

    theme_size = Property(int, _get_size, _set_size)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def keyPressEvent(self, event):
        return

    def mousePressEvent(self, event):
        modifiers = QApplication.keyboardModifiers()
        if modifiers == Qt.AltModifier:
            pos = self.mapToGlobal((self.rect().topLeft()))
            QWhatsThis.showText(pos, self.whatsThis())
        else:
            super(BaseWidget, self).mousePressEvent(event)

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def get_main_layout(self):
        """
        Function that generates the main layout used by the widget
        Override if necessary on new widgets
        :return: QLayout
        """

        layout = QVBoxLayout()
        layout.setContentsMargins(2, 2, 2, 2)
        layout.setSpacing(2)
        layout.setAlignment(Qt.AlignTop)
        return layout

    def ui(self):
        """
        Function that sets up the ui of the widget
        Override it on new widgets (but always call super)
        """

        self.main_layout = self.get_main_layout()
        if self._use_scrollbar:
            layout = QVBoxLayout()
            self.setLayout(layout)
            central_widget = QWidget()
            central_widget.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding))
            scroll = QScrollArea()
            scroll.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
            scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
            scroll.setWidgetResizable(True)
            scroll.setFocusPolicy(Qt.NoFocus)
            layout.addWidget(scroll)
            scroll.setWidget(central_widget)
            central_widget.setLayout(self.main_layout)
            self.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding))
        else:
            self.setLayout(self.main_layout)

    def setup_signals(self):
        """
        Function that set up signals of the widget
        """

        pass

    def set_spacing(self, value):
        """
        Set the spacing used by widget's main layout
        :param value: float
        """

        self.main_layout.setSpacing(value)

    def tiny(self):
        """
        Sets spin box to tiny size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.tiny if widget_theme else theme.Theme.Sizes.TINY

        return self

    def small(self):
        """
        Sets spin box to small size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.small if widget_theme else theme.Theme.Sizes.SMALL

        return self

    def medium(self):
        """
        Sets spin box to medium size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.medium if widget_theme else theme.Theme.Sizes.MEDIUM

        return self

    def large(self):
        """
        Sets spin box to large size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.large if widget_theme else theme.Theme.Sizes.LARGE

        return self

    def huge(self):
        """
        Sets spin box to huge size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.huge if widget_theme else theme.Theme.Sizes.HUGE

        return self


class BaseFrame(QFrame, object):
    mouseReleased = Signal(object)

    def __init__(self, *args, **kwargs):
        super(BaseFrame, self).__init__(*args, **kwargs)

    def mouseReleaseEvent(self, event):
        self.mouseReleased.emit(event)
        return super(BaseFrame, self).mouseReleaseEvent(event)


class ContainerWidget(QWidget, object):
    """
    Basic widget used a
    """

    def __init__(self, parent=None):
        super(ContainerWidget, self).__init__(parent)

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        self.setLayout(layout)

        self.containedWidget = None

    def set_contained_widget(self, widget):
        """
        Sets the current contained widget for this container
        :param widget: QWidget
        """

        self.containedWidget = widget
        if widget:
            widget.setParent(self)
            self.layout().addWidget(widget)

    def clone_and_pass_contained_widget(self):
        """
        Returns a clone of this ContainerWidget
        :return: ContainerWidget
        """

        cloned = ContainerWidget(self.parent())
        cloned.set_contained_widget(self.containedWidget)
        self.set_contained_widget(None)
        return cloned


class DirectoryWidget(BaseWidget, object):
    """
    Widget that contains variables to store current working directory
    """

    def __init__(self, parent=None, **kwargs):
        self.directory = None
        self.last_directory = None
        super(DirectoryWidget, self).__init__(parent=parent, **kwargs)

    def set_directory(self, directory):
        """
        Set the directory used by this widget
        :param directory: str, new directory of the widget
        """

        self.last_directory = self.directory
        self.directory = directory


class PlaceholderWidget(QWidget, object):
    """
    Basic widget that loads custom UI
    """

    def __init__(self, *args):
        super(PlaceholderWidget, self).__init__(*args)
        qtutils.load_widget_ui(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains constants for tpQtLib
"""

from __future__ import print_function, division, absolute_import

__author__ = "Tomas Poveda"
__license__ = "MIT"
__maintainer__ = "Tomas Poveda"
__email__ = "tpovedatd@gmail.com"

# DPI
DEFAULT_DPI = 96

# SPACINGS
SPACING = 2
DEFAULT_SUB_WIDGET_SPACING = 6

# COLORS
DARK_BG_COLOR = tuple([93, 93, 93])
MEDIUM_DARK_BG_COLOR = tuple([73, 73, 73])

# AXISES
AXISES = {
    'x': [255, 0, 0],
    'y': [0, 255, 0],
    'z': [0, 0, 255]
}
